# -*- coding: utf-8 -*-
"""novelWriter GUI Project Settings

 novelWriter – GUI Project Settings
====================================
 Class holding the project settings dialog

 File History:
 Created: 2018-09-29 [0.0.1]

 This file is a part of novelWriter
 Copyright 2018–2020, Veronica Berglyd Olsen

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program. If not, see <https://www.gnu.org/licenses/>.
"""

import nw
import logging

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon, QPixmap, QColor, QBrush
from PyQt5.QtWidgets import (
    QHBoxLayout, QVBoxLayout, QGridLayout, QLineEdit, QPlainTextEdit, QLabel,
    QWidget, QDialogButtonBox, QListWidget, QPushButton, QListWidgetItem,
    QColorDialog, QAbstractItemView, QTreeWidget, QTreeWidgetItem, QComboBox
)

from nw.constants import nwAlert
from nw.gui.custom import QSwitch, PagedDialog, QConfigLayout

logger = logging.getLogger(__name__)

class GuiProjectSettings(PagedDialog):

    def __init__(self, theParent, theProject):
        PagedDialog.__init__(self, theParent)

        logger.debug("Initialising GuiProjectSettings ...")
        self.setObjectName("GuiProjectSettings")

        self.mainConf   = nw.CONFIG
        self.theParent  = theParent
        self.theProject = theProject
        self.optState   = theProject.optState

        self.theProject.countStatus()
        self.setWindowTitle("Project Settings")

        wW = self.mainConf.pxInt(570)
        wH = self.mainConf.pxInt(375)

        self.setMinimumWidth(wW)
        self.setMinimumHeight(wH)
        self.resize(
            self.mainConf.pxInt(self.optState.getInt("GuiProjectSettings", "winWidth",  wW)),
            self.mainConf.pxInt(self.optState.getInt("GuiProjectSettings", "winHeight", wH))
        )

        self.tabMain    = GuiProjectEditMain(self.theParent, self.theProject)
        self.tabMeta    = GuiProjectEditMeta(self.theParent, self.theProject)
        self.tabStatus  = GuiProjectEditStatus(self.theParent, self.theProject, True)
        self.tabImport  = GuiProjectEditStatus(self.theParent, self.theProject, False)
        self.tabReplace = GuiProjectEditReplace(self.theParent, self.theProject)

        self.addTab(self.tabMain,    "Settings")
        self.addTab(self.tabMeta,    "Details")
        self.addTab(self.tabStatus,  "Status")
        self.addTab(self.tabImport,  "Importance")
        self.addTab(self.tabReplace, "Auto-Replace")

        self.buttonBox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        self.buttonBox.accepted.connect(self._doSave)
        self.buttonBox.rejected.connect(self._doClose)
        self.addControls(self.buttonBox)

        logger.debug("GuiProjectSettings initialisation complete")

        return

    ##
    #  Slots
    ##

    def _doSave(self):
        """Save settings and close dialog.
        """
        logger.verbose("GuiProjectSettings save button clicked")

        projName    = self.tabMain.editName.text()
        bookTitle   = self.tabMain.editTitle.text()
        bookAuthors = self.tabMain.editAuthors.toPlainText()
        spellLang   = self.tabMain.spellLang.currentData()
        doBackup    = not self.tabMain.doBackup.isChecked()

        self.theProject.setProjectName(projName)
        self.theProject.setBookTitle(bookTitle)
        self.theProject.setBookAuthors(bookAuthors)
        self.theProject.setSpellLang(spellLang)
        self.theProject.setProjBackup(doBackup)

        if self.tabStatus.colChanged:
            statusCol = self.tabStatus.getNewList()
            self.theProject.setStatusColours(statusCol)

        if self.tabImport.colChanged:
            importCol = self.tabImport.getNewList()
            self.theProject.setImportColours(importCol)

        if self.tabStatus.colChanged or self.tabImport.colChanged:
            self.theParent.rebuildTree()

        if self.tabReplace.arChanged:
            newList = self.tabReplace.getNewList()
            self.theProject.setAutoReplace(newList)

        self._saveGuiSettings()
        self.accept()

        return

    def _doClose(self):
        """Save settings and close the dialog.
        """
        self._saveGuiSettings()
        self.reject()
        return

    ##
    #  Internal Functions
    ##

    def _saveGuiSettings(self):
        """Save GUI settings.
        """
        winWidth    = self.mainConf.rpxInt(self.width())
        winHeight   = self.mainConf.rpxInt(self.height())
        replaceColW = self.mainConf.rpxInt(self.tabReplace.listBox.columnWidth(0))

        self.optState.setValue("GuiProjectSettings", "winWidth",    winWidth)
        self.optState.setValue("GuiProjectSettings", "winHeight",   winHeight)
        self.optState.setValue("GuiProjectSettings", "replaceColW", replaceColW)

        return

# END Class GuiProjectSettings

class GuiProjectEditMain(QWidget):

    def __init__(self, theParent, theProject):
        QWidget.__init__(self, theParent)

        self.mainConf   = nw.CONFIG
        self.theParent  = theParent
        self.theProject = theProject

        # The Form
        self.mainForm = QConfigLayout()
        self.mainForm.setHelpTextStyle(self.theParent.theTheme.helpText)
        self.setLayout(self.mainForm)

        self.mainForm.addGroupLabel("Project Settings")

        xW = self.mainConf.pxInt(250)
        xH = self.mainConf.pxInt(100)

        self.editName = QLineEdit()
        self.editName.setMaxLength(200)
        self.editName.setFixedWidth(xW)
        self.editName.setText(self.theProject.projName)
        self.mainForm.addRow(
            "Working title",
            self.editName,
            "Should be set only once."
        )

        self.editTitle = QLineEdit()
        self.editTitle.setMaxLength(200)
        self.editTitle.setFixedWidth(xW)
        self.editTitle.setText(self.theProject.bookTitle)
        self.mainForm.addRow(
            "Novel title",
            self.editTitle,
            "Change whenever you want!"
        )

        self.editAuthors = QPlainTextEdit()
        bookAuthors = ""
        for bookAuthor in self.theProject.bookAuthors:
            bookAuthors += bookAuthor+"\n"
        self.editAuthors.setPlainText(bookAuthors)
        self.editAuthors.setFixedHeight(xH)
        self.editAuthors.setFixedWidth(xW)
        self.mainForm.addRow(
            "Author(s)",
            self.editAuthors,
            "One name per line."
        )

        self.spellLang = QComboBox(self)
        theDict = self.theParent.docEditor.theDict
        self.spellLang.addItem("Default", "None")
        if theDict is not None:
            for spTag, spName in theDict.listDictionaries():
                self.spellLang.addItem(spName, spTag)

        self.mainForm.addRow(
            "Spell check language",
            self.spellLang,
            "Overrides main preferences."
        )

        if self.theProject.projLang is None:
            spellIdx = 0
        else:
            spellIdx = self.spellLang.findData(self.theProject.projLang)
        if spellIdx != -1:
            self.spellLang.setCurrentIndex(spellIdx)

        self.doBackup = QSwitch(self)
        self.doBackup.setChecked(not self.theProject.doBackup)
        self.mainForm.addRow(
            "No backup on close",
            self.doBackup,
            "Overrides main preferences."
        )

        return

# END Class GuiProjectEditMain

class GuiProjectEditMeta(QWidget):

    def __init__(self, theParent, theProject):
        QWidget.__init__(self, theParent)

        self.mainConf   = nw.CONFIG
        self.theParent  = theParent
        self.theProject = theProject

        xInd = self.mainConf.pxInt(8)

        # The Form
        self.mainForm = QGridLayout()
        self.setLayout(self.mainForm)

        self.headLabel = QLabel("<b>Project Details</b>")

        self.nameLabel = QLabel("Working title:")
        self.nameLabel.setIndent(xInd)
        self.nameValue = QLabel(self.theProject.projName)
        self.nameValue.setWordWrap(True)

        self.pathLabel = QLabel("Project path:")
        self.pathLabel.setIndent(xInd)
        self.pathValue = QLabel(self.theProject.projPath)
        self.pathValue.setWordWrap(True)
        self.pathValue.setTextInteractionFlags(Qt.TextSelectableByMouse)
        self.pathValue.setCursor(Qt.IBeamCursor)

        self.revLabel = QLabel("Revision count:")
        self.revLabel.setIndent(xInd)
        self.revValue = QLabel(f"{self.theProject.saveCount:n}")

        editHours = self.theProject.editTime/3600
        self.editLabel = QLabel("Edit time:")
        self.editLabel.setIndent(xInd)
        self.editValue = QLabel(f"{editHours:.2f} hours")

        self.statsLabel = QLabel("<b>Project Stats</b>")

        nR, nD, nF = self.theProject.projTree.countTypes()

        self.nRootLabel = QLabel("Root folders:")
        self.nRootLabel.setIndent(xInd)
        self.nRootValue = QLabel(f"{nR:n}")

        self.nDirLabel = QLabel("Folders:")
        self.nDirLabel.setIndent(xInd)
        self.nDirValue = QLabel(f"{nD:n}")

        self.nFileLabel = QLabel("Documents:")
        self.nFileLabel.setIndent(xInd)
        self.nFileValue = QLabel(f"{nF:n}")

        self.wordsLabel = QLabel("Word count:")
        self.wordsLabel.setIndent(xInd)
        self.wordsValue = QLabel(f"{self.theProject.currWCount:n}")

        self.mainForm.addWidget(self.headLabel,  0, 0, 1, 2, Qt.AlignTop)
        self.mainForm.addWidget(self.nameLabel,  1, 0, 1, 1, Qt.AlignTop)
        self.mainForm.addWidget(self.nameValue,  1, 1, 1, 1, Qt.AlignTop)
        self.mainForm.addWidget(self.pathLabel,  2, 0, 1, 1, Qt.AlignTop)
        self.mainForm.addWidget(self.pathValue,  2, 1, 1, 1, Qt.AlignTop)
        self.mainForm.addWidget(self.revLabel,   3, 0, 1, 1, Qt.AlignTop)
        self.mainForm.addWidget(self.revValue,   3, 1, 1, 1, Qt.AlignTop)
        self.mainForm.addWidget(self.editLabel,  4, 0, 1, 1, Qt.AlignTop)
        self.mainForm.addWidget(self.editValue,  4, 1, 1, 1, Qt.AlignTop)

        self.mainForm.addWidget(self.statsLabel, 5, 0, 1, 2, Qt.AlignTop)
        self.mainForm.addWidget(self.nRootLabel, 6, 0, 1, 1, Qt.AlignTop)
        self.mainForm.addWidget(self.nRootValue, 6, 1, 1, 1, Qt.AlignTop)
        self.mainForm.addWidget(self.nDirLabel,  7, 0, 1, 1, Qt.AlignTop)
        self.mainForm.addWidget(self.nDirValue,  7, 1, 1, 1, Qt.AlignTop)
        self.mainForm.addWidget(self.nFileLabel, 8, 0, 1, 1, Qt.AlignTop)
        self.mainForm.addWidget(self.nFileValue, 8, 1, 1, 1, Qt.AlignTop)
        self.mainForm.addWidget(self.wordsLabel, 9, 0, 1, 1, Qt.AlignTop)
        self.mainForm.addWidget(self.wordsValue, 9, 1, 1, 1, Qt.AlignTop)

        self.mainForm.setVerticalSpacing(self.mainConf.pxInt(6))
        self.mainForm.setHorizontalSpacing(self.mainConf.pxInt(12))
        self.mainForm.setColumnStretch(0, 0)
        self.mainForm.setColumnStretch(1, 1)
        self.mainForm.setRowStretch(10, 1)

        return

# END Class GuiProjectEditMeta

class GuiProjectEditStatus(QWidget):

    def __init__(self, theParent, theProject, isStatus):
        QWidget.__init__(self, theParent)

        self.mainConf   = nw.CONFIG
        self.theParent  = theParent
        self.theProject = theProject
        self.theTheme   = theParent.theTheme
        if isStatus:
            self.theStatus = self.theProject.statusItems
        else:
            self.theStatus = self.theProject.importItems

        self.colData    = []
        self.colCounts  = []
        self.colChanged = False
        self.selColour  = None

        self.iPx = self.theTheme.baseIconSize

        self.outerBox = QVBoxLayout()
        self.mainBox  = QHBoxLayout()
        self.mainForm = QVBoxLayout()

        self.listBox = QListWidget()
        self.listBox.setDragDropMode(QAbstractItemView.InternalMove)
        self.listBox.itemSelectionChanged.connect(self._selectedItem)
        self.listBox.model().rowsMoved.connect(self._rowsMoved)

        for iName, iCol, nUse in self.theStatus:
            self._addItem(iName, iCol, iName, nUse)

        self.editName = QLineEdit()
        self.editName.setMaxLength(40)
        self.editName.setEnabled(False)
        self.newButton  = QPushButton("New")
        self.delButton  = QPushButton("Delete")
        self.saveButton = QPushButton("Save")
        self.colPixmap  = QPixmap(self.iPx, self.iPx)
        self.colPixmap.fill(QColor(120, 120, 120))
        self.colButton  = QPushButton(QIcon(self.colPixmap), "Colour")
        self.colButton.setIconSize(self.colPixmap.rect().size())

        self.newButton.clicked.connect(self._newItem)
        self.delButton.clicked.connect(self._delItem)
        self.saveButton.clicked.connect(self._saveItem)
        self.colButton.clicked.connect(self._selectColour)

        self.mainForm.addWidget(self.newButton)
        self.mainForm.addWidget(self.delButton)
        self.mainForm.addStretch(1)
        self.mainForm.addWidget(QLabel("<b>Name</b>"))
        self.mainForm.addWidget(self.editName)
        self.mainForm.addWidget(self.colButton)
        self.mainForm.addStretch(1)
        self.mainForm.addWidget(self.saveButton)

        self.mainBox.addWidget(self.listBox)
        self.mainBox.addLayout(self.mainForm)

        if isStatus:
            self.outerBox.addWidget(QLabel("<b>Novel File Status Levels</b>"))
        else:
            self.outerBox.addWidget(QLabel("<b>Note File Importance Levels</b>"))
        self.outerBox.addLayout(self.mainBox)

        self.setLayout(self.outerBox)

        return

    def getNewList(self):
        if self.colChanged:
            newList = []
            for n in range(self.listBox.count()):
                nItem = self.listBox.item(n)
                nIdx  = nItem.data(Qt.UserRole)
                newList.append(self.colData[nIdx])
            return newList
        return None

    ##
    #  User Actions
    ##

    def _selectColour(self):
        """Open a dialog to select the status icon colour.
        """
        logger.verbose("Item colour button clicked")
        if self.selColour is not None:
            newCol = QColorDialog.getColor(
                self.selColour, self, "Select Colour", QColorDialog.DontUseNativeDialog
            )
            if newCol.isValid():
                self.selColour = newCol
                colPixmap = QPixmap(self.iPx, self.iPx)
                colPixmap.fill(newCol)
                self.colButton.setIcon(QIcon(colPixmap))
                self.colButton.setIconSize(colPixmap.rect().size())
        return

    def _newItem(self):
        """Create a new status item.
        """
        logger.verbose("New item button clicked")
        newItem = self._addItem("New Item", (0, 0, 0), None, 0)
        newItem.setBackground(QBrush(QColor(0, 255, 0, 80)))
        self.colChanged = True
        return

    def _delItem(self):
        """Delete a status item.
        """
        logger.verbose("Delete item button clicked")
        selItem = self._getSelectedItem()
        if selItem is not None:
            iRow   = self.listBox.row(selItem)
            selIdx = selItem.data(Qt.UserRole)
            if self.colCounts[selIdx] == 0:
                self.listBox.takeItem(iRow)
                self.colChanged = True
            else:
                self.theParent.makeAlert(
                    "Cannot delete status item that is in use.", nwAlert.ERROR
                )
        return

    def _saveItem(self):
        """Save changes made to a status item.
        """
        logger.verbose("Save item button clicked")
        selItem = self._getSelectedItem()
        if selItem is not None:
            selIdx = selItem.data(Qt.UserRole)
            self.colData[selIdx] = (
                self.editName.text().strip(),
                self.selColour.red(),
                self.selColour.green(),
                self.selColour.blue(),
                self.colData[selIdx][4]
            )
            selItem.setText("%s [%d]" % (self.colData[selIdx][0], self.colCounts[selIdx]))
            selItem.setIcon(self.colButton.icon())
            self.editName.setEnabled(False)
            self.colChanged = True
        return

    def _addItem(self, iName, iCol, oName, nUse):
        """Add a status item to the list.
        """
        newIcon = QPixmap(self.iPx, self.iPx)
        newIcon.fill(QColor(*iCol))
        newItem = QListWidgetItem()
        newItem.setText("%s [%d]" % (iName, nUse))
        newItem.setIcon(QIcon(newIcon))
        newItem.setData(Qt.UserRole, len(self.colData))
        self.listBox.addItem(newItem)
        self.colData.append((iName, iCol[0], iCol[1], iCol[2], oName))
        self.colCounts.append(nUse)
        return newItem

    def _selectedItem(self):
        """Extract the info of a selected item and populate the settings
        boxes and button.
        """
        logger.verbose("Item selected")
        selItem = self._getSelectedItem()
        if selItem is not None:
            selIdx = selItem.data(Qt.UserRole)
            selVal = self.colData[selIdx]
            self.selColour = QColor(selVal[1], selVal[2], selVal[3])
            newIcon = QPixmap(self.iPx, self.iPx)
            newIcon.fill(self.selColour)
            self.editName.setText(selVal[0])
            self.colButton.setIcon(QIcon(newIcon))
            self.editName.setEnabled(True)
            self.editName.selectAll()
            self.editName.setFocus()
        return

    ##
    #  Internal Functions
    ##

    def _getSelectedItem(self):
        """Get the currently selected item.
        """
        selItem = self.listBox.selectedItems()
        if len(selItem) == 0:
            return None
        if isinstance(selItem[0], QListWidgetItem):
            return selItem[0]
        return None

    def _rowsMoved(self):
        """A row has been moved, so sett the changed flag.
        """
        logger.verbose("A drag move event occurred")
        self.colChanged = True
        return

# END Class GuiProjectEditStatus

class GuiProjectEditReplace(QWidget):

    def __init__(self, theParent, theProject):
        QWidget.__init__(self, theParent)

        self.mainConf   = nw.CONFIG
        self.theParent  = theParent
        self.theTheme   = theParent.theTheme
        self.theProject = theProject
        self.optState   = theProject.optState
        self.arChanged  = False

        self.outerBox  = QVBoxLayout()
        self.bottomBox = QHBoxLayout()

        wCol0 = self.mainConf.pxInt(
            self.optState.getInt("GuiProjectSettings", "replaceColW", 100)
        )
        self.listBox = QTreeWidget()
        self.listBox.setHeaderLabels(["Keyword", "Replace With"])
        self.listBox.itemSelectionChanged.connect(self._selectedItem)
        self.listBox.setColumnWidth(0, wCol0)
        self.listBox.setIndentation(0)

        for aKey, aVal in self.theProject.autoReplace.items():
            newItem = QTreeWidgetItem(["<%s>" % aKey, aVal])
            self.listBox.addTopLevelItem(newItem)

        self.listBox.sortByColumn(0, Qt.AscendingOrder)
        self.listBox.setSortingEnabled(True)

        self.editKey    = QLineEdit()
        self.editValue  = QLineEdit()
        self.saveButton = QPushButton(self.theTheme.getIcon("done"), "")
        self.addButton  = QPushButton(self.theTheme.getIcon("add"), "")
        self.delButton  = QPushButton(self.theTheme.getIcon("remove"), "")
        self.saveButton.setToolTip("Save entry")
        self.addButton.setToolTip("Add new entry")
        self.delButton.setToolTip("Delete selected entry")

        self.editKey.setEnabled(False)
        self.editKey.setMaxLength(40)
        self.editValue.setEnabled(False)
        self.editValue.setMaxLength(80)

        self.saveButton.clicked.connect(self._saveEntry)
        self.addButton.clicked.connect(self._addEntry)
        self.delButton.clicked.connect(self._delEntry)

        self.bottomBox.addWidget(self.editKey, 2)
        self.bottomBox.addWidget(self.editValue, 3)
        self.bottomBox.addWidget(self.saveButton)
        self.bottomBox.addWidget(self.addButton)
        self.bottomBox.addWidget(self.delButton)

        self.outerBox.addWidget(QLabel("<b>Text Replace List for Preview and Export</b>"))
        self.outerBox.addWidget(self.listBox)
        self.outerBox.addLayout(self.bottomBox)
        self.setLayout(self.outerBox)

        return

    def getNewList(self):
        """Extract the list from the widget.
        """
        newList = {}
        for n in range(self.listBox.topLevelItemCount()):
            tItem = self.listBox.topLevelItem(n)
            aKey = self._stripNotAllowed(tItem.text(0))
            aVal = tItem.text(1)
            if len(aKey) > 0:
                newList[aKey] = aVal
        return newList

    ##
    #  Internal Functions
    ##

    def _selectedItem(self):
        """Extract the details from the selected item and populate the
        edit form.
        """
        selItem = self._getSelectedItem()
        if selItem is None:
            return False
        editKey = self._stripNotAllowed(selItem.text(0))
        editVal = selItem.text(1)
        self.editKey.setText(editKey)
        self.editValue.setText(editVal)
        self.editKey.setEnabled(True)
        self.editValue.setEnabled(True)
        self.editKey.selectAll()
        self.editKey.setFocus()
        return True

    def _saveEntry(self):
        """Save the form data into the list widget.
        """
        selItem = self._getSelectedItem()
        if selItem is None:
            return False

        newKey  = self.editKey.text()
        newVal  = self.editValue.text()
        saveKey = self._stripNotAllowed(newKey)

        if len(saveKey) > 0 and len(newVal) > 0:
            selItem.setText(0, "<%s>" % saveKey)
            selItem.setText(1, newVal)
            self.editKey.clear()
            self.editValue.clear()
            self.editKey.setEnabled(False)
            self.editValue.setEnabled(False)
            self.listBox.clearSelection()
            self.arChanged = True

        return

    def _addEntry(self):
        """Add a new list entry.
        """
        saveKey = "<keyword%d>" % (self.listBox.topLevelItemCount() + 1)
        newVal  = ""
        newItem = QTreeWidgetItem([saveKey, newVal])
        self.listBox.addTopLevelItem(newItem)
        return True

    def _delEntry(self):
        """Delete the selected entry.
        """
        selItem = self._getSelectedItem()
        if selItem is None:
            return False
        self.listBox.takeTopLevelItem(self.listBox.indexOfTopLevelItem(selItem))
        self.arChanged = True
        return True

    def _getSelectedItem(self):
        """Extract the currently selected item.
        """
        selItem = self.listBox.selectedItems()
        if len(selItem) == 0:
            return None
        return selItem[0]

    def _stripNotAllowed(self, theKey):
        """Clean up the replace key string.
        """
        retKey = ""
        for c in theKey:
            if c.isalnum():
                retKey += c
        return retKey

# END Class GuiProjectEditReplace



# -*- coding: utf-8 -*-
"""novelWriter HTML Text Converter

 novelWriter – HTML Text Converter
===================================
 Extends the Tokenizer class to write HTML

 File History:
 Created: 2019-05-07 [0.0.1]

 This file is a part of novelWriter
 Copyright 2018–2020, Veronica Berglyd Olsen

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program. If not, see <https://www.gnu.org/licenses/>.
"""

import logging
import re

from nw.core.tokenizer import Tokenizer
from nw.constants import nwUnicode, nwLabels, nwKeyWords

logger = logging.getLogger(__name__)

class ToHtml(Tokenizer):

    M_PREVIEW = 0 # Tweak output for the DocViewer
    M_EXPORT  = 1 # Tweak output for saving to HTML or printing
    M_EBOOK   = 2 # Tweak output for converting to epub

    def __init__(self, theProject, theParent):
        Tokenizer.__init__(self, theProject, theParent)

        self.genMode = self.M_EXPORT
        self.cssStyles = True

        self.repDict = {
            "<" : "&lt;",
            ">" : "&gt;",
            "&" : "&amp;",
            nwUnicode.U_ENDASH : nwUnicode.H_ENDASH,
            nwUnicode.U_EMDASH : nwUnicode.H_EMDASH,
            nwUnicode.U_HELLIP : nwUnicode.H_HELLIP,
            nwUnicode.U_NBSP   : nwUnicode.H_NBSP,
            nwUnicode.U_THNSP  : nwUnicode.H_THNSP,
            nwUnicode.U_THNBSP : nwUnicode.H_THNBSP,
            nwUnicode.U_MAPOSS : nwUnicode.H_RSQUO,
        }
        self.revDict = {}
        self.reReplace = []
        self.reReverse = []
        self._buildRegEx()

        return

    ##
    #  Setters
    ##

    def setPreview(self, forPreview, doComments, doSynopsis):
        """If we're using this class to generate markdown preview, we
        need to make a few changes to formatting, which is managed by
        these flags.
        """
        if forPreview:
            self.genMode    = self.M_PREVIEW
            self.doKeywords = True
            self.doComments = doComments
            self.doSynopsis = doSynopsis
        return

    def setStyles(self, cssStyles):
        """Enable/disable CSS styling. Some elements may still have
        class tags.
        """
        self.cssStyles = cssStyles
        return

    ##
    #  Class Methods
    ##

    def doAutoReplace(self):
        """Extend the auto-replace to also properly encode some unicode
        characters into their respective HTML entities.
        """
        Tokenizer.doAutoReplace(self)
        self.theText = self.reReplace.sub(
            lambda x: self.repDict[x.group(0)], self.theText
        )
        return

    def doPostProcessing(self):
        """Reverse the html entities replacement on the markdown text.
        Otherwise, all the &something; bits will also be in there.
        """
        Tokenizer.doPostProcessing(self)
        if self.genMode == self.M_PREVIEW:
            # Doesn't matter for preview as we don't use the markdown
            return
        self.theMarkdown = self.reReverse.sub(
            lambda x: self.revDict[x.group(0)], self.theMarkdown
        )
        return

    def doConvert(self):
        """Convert the list of text tokens into a HTML document saved
        to theResult.
        """
        if self.genMode == self.M_PREVIEW:
            htmlTags = {     # HTML4 + CSS2
                self.FMT_B_B : "<b>",
                self.FMT_B_E : "</b>",
                self.FMT_I_B : "<i>",
                self.FMT_I_E : "</i>",
                self.FMT_D_B : "<span style='text-decoration: line-through;'>",
                self.FMT_D_E : "</span>",
            }
        else:
            htmlTags = {     # HTML5
                self.FMT_B_B : "<strong>",
                self.FMT_B_E : "</strong>",
                self.FMT_I_B : "<em>",
                self.FMT_I_E : "</em>",
                self.FMT_D_B : "<del>",
                self.FMT_D_E : "</del>",
            }

        if self.isNovel and self.genMode != self.M_PREVIEW:
            # For novel files for export, we bump the titles one level
            # up as this is more useful for printing and word processor
            # imports.
            h1 = "h1 class='title'"
            h2 = "h1"
            h3 = "h2"
            h4 = "h3"
        else:
            h1 = "h1"
            h2 = "h2"
            h3 = "h3"
            h4 = "h4"

        self.theResult = ""

        thisPar = []
        parStyle = None
        tmpResult = []
        hasHardBreak = False
        for tType, tLine, tText, tFormat, tStyle in self.theTokens:

            # Styles
            aStyle = []
            if tStyle is not None and self.cssStyles:
                if tStyle & self.A_LEFT:
                    aStyle.append("text-align: left;")
                if tStyle & self.A_RIGHT:
                    aStyle.append("text-align: right;")
                if tStyle & self.A_CENTRE:
                    aStyle.append("text-align: center;")
                if tStyle & self.A_JUSTIFY:
                    aStyle.append("text-align: justify;")
                if tStyle & self.A_PBB:
                    aStyle.append("page-break-before: always;")
                if tStyle & self.A_PBB_AV:
                    aStyle.append("page-break-before: avoid;")
                if tStyle & self.A_PBB_NO:
                    aStyle.append("page-break-before: never;")
                if tStyle & self.A_PBA:
                    aStyle.append("page-break-after: always;")
                if tStyle & self.A_PBA_AV:
                    aStyle.append("page-break-after: avoid;")
                if tStyle & self.A_PBA_NO:
                    aStyle.append("page-break-after: never;")

            if len(aStyle) > 0:
                hStyle = " style='%s'" % (" ".join(aStyle))
            else:
                hStyle = ""

            if self.linkHeaders:
                aNm = "<a name='T%06d'></a>" % tLine
            else:
                aNm = ""

            # Process TextType
            if tType == self.T_EMPTY:
                if parStyle is None:
                    parStyle = ""
                if hasHardBreak and self.cssStyles:
                    parClass = " class='break'"
                else:
                    parClass = ""
                if len(thisPar) > 0:
                    tTemp = "".join(thisPar)
                    tmpResult.append("<p%s%s>%s</p>\n" % (parStyle, parClass, tTemp.rstrip()))
                thisPar = []
                parStyle = None
                hasHardBreak = False

            elif tType == self.T_TITLE:
                tHead = tText.replace(r"\\", "<br/>")
                tmpResult.append("<h1 class='title'%s>%s%s</h1>\n" % (hStyle, aNm, tHead))

            elif tType == self.T_HEAD1:
                tHead = tText.replace(r"\\", "<br/>")
                tmpResult.append("<%s%s>%s%s</%s>\n" % (h1, hStyle, aNm, tHead, h1))

            elif tType == self.T_HEAD2:
                tHead = tText.replace(r"\\", "<br/>")
                tmpResult.append("<%s%s>%s%s</%s>\n" % (h2, hStyle, aNm, tHead, h2))

            elif tType == self.T_HEAD3:
                tHead = tText.replace(r"\\", "<br/>")
                tmpResult.append("<%s%s>%s%s</%s>\n" % (h3, hStyle, aNm, tHead, h3))

            elif tType == self.T_HEAD4:
                tHead = tText.replace(r"\\", "<br/>")
                tmpResult.append("<%s%s>%s%s</%s>\n" % (h4, hStyle, aNm, tHead, h4))

            elif tType == self.T_SEP:
                tmpResult.append("<p class='sep'>%s</p>\n" % tText)

            elif tType == self.T_SKIP:
                tmpResult.append("<p class='skip'>&nbsp;</p>\n")

            elif tType == self.T_TEXT:
                tTemp = tText
                if parStyle is None:
                    parStyle = hStyle
                for xPos, xLen, xFmt in reversed(tFormat):
                    tTemp = tTemp[:xPos]+htmlTags[xFmt]+tTemp[xPos+xLen:]
                if tText.endswith("  "):
                    thisPar.append(tTemp.rstrip()+"<br/>")
                    hasHardBreak = True
                else:
                    thisPar.append(tTemp.rstrip()+" ")

            elif tType == self.T_SYNOPSIS and self.doSynopsis:
                tmpResult.append(self._formatSynopsis(tText))

            elif tType == self.T_COMMENT and self.doComments:
                tmpResult.append(self._formatComments(tText))

            elif tType == self.T_KEYWORD and self.doKeywords:
                tmpResult.append(self._formatKeywords(tText))

        self.theResult = "".join(tmpResult)
        tmpResult = []

        return

    def getStyleSheet(self):
        """Generate a stylesheet appropriate for the current settings.
        """
        theStyles = []
        if not self.cssStyles:
            return theStyles

        if self.doJustify:
            theStyles.append(r"p {text-align: justify;}")
        else:
            theStyles.append(r"p {text-align: left;}")

        theStyles.append(r"h1, h2 {color: rgb(66, 113, 174);}")
        theStyles.append(r"h3, h4 {color: rgb(50, 50, 50);}")
        theStyles.append(r"h1, h2, h3, h4 {page-break-after: avoid;}")
        theStyles.append(r"a {color: rgb(66, 113, 174);}")
        theStyles.append(r".title {font-size: 2.5em;}")
        theStyles.append(r".tags {color: rgb(245, 135, 31); font-weight: bold;}")
        theStyles.append(r".break {text-align: left;}")
        theStyles.append(r".sep {text-align: center; margin-top: 1em; margin-bottom: 1em;}")
        theStyles.append(r".skip {margin-top: 1em; margin-bottom: 1em;}")
        theStyles.append(r".synopsis {font-style: italic;}")
        theStyles.append(r".comment {font-style: italic; color: rgb(100, 100, 100);}")

        return theStyles

    ##
    #  Internal Functions
    ##

    def _formatSynopsis(self, tText):
        """Apply HTML formatting to synopsis.
        """
        if self.genMode == self.M_PREVIEW:
            return "<p class='comment'><span class='synopsis'>Synopsis: </span>%s</p>\n" % tText
        else:
            return "<p class='synopsis'><strong>Synopsis: </strong>%s</p>\n" % tText

    def _formatComments(self, tText):
        """Apply HTML formatting to comments.
        """
        if self.genMode == self.M_PREVIEW:
            return "<p class='comment'>%s</p>\n" % tText
        else:
            return "<p class='comment'><strong>Comment: </strong>%s</p>\n" % tText

    def _formatKeywords(self, tText):
        """Apply HTML formatting to keywords.
        """
        isValid, theBits, thePos = self.theParent.theIndex.scanThis("@"+tText)
        if not isValid or not theBits:
            return ""

        retText = ""
        refTags = []
        if theBits[0] in nwLabels.KEY_NAME:
            retText += "<span class='tags'>%s:</span> " % nwLabels.KEY_NAME[theBits[0]]
            if len(theBits) > 1:
                if theBits[0] == nwKeyWords.TAG_KEY:
                    retText += "<a name='tag_%s'>%s</a>" % (
                        theBits[1], theBits[1]
                    )
                else:
                    if self.genMode == self.M_PREVIEW:
                        for tTag in theBits[1:]:
                            refTags.append("<a href='#%s=%s'>%s</a>" % (
                                theBits[0][1:], tTag, tTag
                            ))
                        retText += ", ".join(refTags)
                    else:
                        for tTag in theBits[1:]:
                            refTags.append("<a href='#tag_%s'>%s</a>" % (
                                tTag, tTag
                            ))
                        retText += ", ".join(refTags)

        return "<div>%s</div>" % retText

    def _buildRegEx(self):
        """Build the regular expressions
        """
        self.revDict = dict(map(reversed, self.repDict.items()))
        self.reReplace = re.compile(
            "|".join([re.escape(k) for k in self.repDict.keys()]), flags=re.DOTALL
        )
        self.reReverse = re.compile(
            "|".join([re.escape(k) for k in self.revDict.keys()]), flags=re.DOTALL
        )
        return

# END Class ToHtml



# -*- coding: utf-8 -*-
"""novelWriter ISO Codes

 novelWriter – ISO Codes
=========================
 Handles translating language codes to language names

 File History:
 Created: 2019-11-05 [0.4.0]

 This file is a part of novelWriter
 Copyright 2018–2020, Veronica Berglyd Olsen

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program. If not, see <https://www.gnu.org/licenses/>.
"""

class isoLanguage():

    ISO_639_1 = {
        "aa" : "Afar",
        "ab" : "Abkhazian",
        "ae" : "Avestan",
        "af" : "Afrikaans",
        "ak" : "Akan",
        "am" : "Amharic",
        "an" : "Aragonese",
        "ar" : "Arabic",
        "as" : "Assamese",
        "av" : "Avaric",
        "ay" : "Aymara",
        "az" : "Azerbaijani",
        "ba" : "Bashkir",
        "be" : "Belarusian",
        "bg" : "Bulgarian",
        "bh" : "Bihari languages",
        "bi" : "Bislama",
        "bm" : "Bambara",
        "bn" : "Bengali",
        "bo" : "Tibetan",
        "br" : "Breton",
        "bs" : "Bosnian",
        "ca" : "Catalan",
        "ce" : "Chechen",
        "ch" : "Chamorro",
        "co" : "Corsican",
        "cr" : "Cree",
        "cs" : "Czech",
        "cu" : "Church Slavic",
        "cv" : "Chuvash",
        "cy" : "Welsh",
        "da" : "Danish",
        "de" : "German",
        "dv" : "Divehi",
        "dz" : "Dzongkha",
        "ee" : "Ewe",
        "el" : "Modern Greek",
        "en" : "English",
        "eo" : "Esperanto",
        "es" : "Spanish",
        "et" : "Estonian",
        "eu" : "Basque",
        "fa" : "Persian",
        "ff" : "Fulah",
        "fi" : "Finnish",
        "fj" : "Fijian",
        "fo" : "Faroese",
        "fr" : "French",
        "fy" : "Western Frisian",
        "ga" : "Irish",
        "gd" : "Gaelic",
        "gl" : "Galician",
        "gn" : "Guarani",
        "gu" : "Gujarati",
        "gv" : "Manx",
        "ha" : "Hausa",
        "he" : "Hebrew",
        "hi" : "Hindi",
        "ho" : "Hiri Motu",
        "hr" : "Croatian",
        "ht" : "Haitian",
        "hu" : "Hungarian",
        "hy" : "Armenian",
        "hz" : "Herero",
        "ia" : "Interlingua",
        "id" : "Indonesian",
        "ie" : "Interlingue",
        "ig" : "Igbo",
        "ii" : "Sichuan Yi",
        "ik" : "Inupiaq",
        "io" : "Ido",
        "is" : "Icelandic",
        "it" : "Italian",
        "iu" : "Inuktitut",
        "ja" : "Japanese",
        "jv" : "Javanese",
        "ka" : "Georgian",
        "kg" : "Kongo",
        "ki" : "Kikuyu",
        "kj" : "Kuanyama",
        "kk" : "Kazakh",
        "kl" : "Kalaallisut",
        "km" : "Central Khmer",
        "kn" : "Kannada",
        "ko" : "Korean",
        "kr" : "Kanuri",
        "ks" : "Kashmiri",
        "ku" : "Kurdish",
        "kv" : "Komi",
        "kw" : "Cornish",
        "ky" : "Kirghiz",
        "la" : "Latin",
        "lb" : "Luxembourgish",
        "lg" : "Ganda",
        "li" : "Limburgan",
        "ln" : "Lingala",
        "lo" : "Lao",
        "lt" : "Lithuanian",
        "lu" : "Luba-Katanga",
        "lv" : "Latvian",
        "mg" : "Malagasy",
        "mh" : "Marshallese",
        "mi" : "Maori",
        "mk" : "Macedonian",
        "ml" : "Malayalam",
        "mn" : "Mongolian",
        "mr" : "Marathi",
        "ms" : "Malay",
        "mt" : "Maltese",
        "my" : "Burmese",
        "na" : "Nauru",
        "nb" : "Norwegian Bokmål",
        "nd" : "North Ndebele",
        "ne" : "Nepali",
        "ng" : "Ndonga",
        "nl" : "Dutch",
        "nn" : "Norwegian Nynorsk",
        "no" : "Norwegian",
        "nr" : "South Ndebele",
        "nv" : "Navajo",
        "ny" : "Chichewa",
        "oc" : "Occitan",
        "oj" : "Ojibwa",
        "om" : "Oromo",
        "or" : "Oriya",
        "os" : "Ossetian",
        "pa" : "Panjabi",
        "pi" : "Pali",
        "pl" : "Polish",
        "ps" : "Pushto",
        "pt" : "Portuguese",
        "qu" : "Quechua",
        "rm" : "Romansh",
        "rn" : "Rundi",
        "ro" : "Romanian",
        "ru" : "Russian",
        "rw" : "Kinyarwanda",
        "sa" : "Sanskrit",
        "sc" : "Sardinian",
        "sd" : "Sindhi",
        "se" : "Northern Sami",
        "sg" : "Sango",
        "si" : "Sinhala",
        "sk" : "Slovak",
        "sl" : "Slovenian",
        "sm" : "Samoan",
        "sn" : "Shona",
        "so" : "Somali",
        "sq" : "Albanian",
        "sr" : "Serbian",
        "ss" : "Swati",
        "st" : "Southern Sotho",
        "su" : "Sundanese",
        "sv" : "Swedish",
        "sw" : "Swahili",
        "ta" : "Tamil",
        "te" : "Telugu",
        "tg" : "Tajik",
        "th" : "Thai",
        "ti" : "Tigrinya",
        "tk" : "Turkmen",
        "tl" : "Tagalog",
        "tn" : "Tswana",
        "to" : "Tonga",
        "tr" : "Turkish",
        "ts" : "Tsonga",
        "tt" : "Tatar",
        "tw" : "Twi",
        "ty" : "Tahitian",
        "ug" : "Uighur",
        "uk" : "Ukrainian",
        "ur" : "Urdu",
        "uz" : "Uzbek",
        "ve" : "Venda",
        "vi" : "Vietnamese",
        "vo" : "Volapük",
        "wa" : "Walloon",
        "wo" : "Wolof",
        "xh" : "Xhosa",
        "yi" : "Yiddish",
        "yo" : "Yoruba",
        "za" : "Zhuang",
        "zh" : "Chinese",
        "zu" : "Zulu",
    }

# END Class isoLanguage

class isoCountry():

    ISO_3166_1_alpha_2 = {
        "AD" : "Andorra",
        "AE" : "United Arab Emirates",
        "AF" : "Afghanistan",
        "AG" : "Antigua and Barbuda",
        "AI" : "Anguilla",
        "AL" : "Albania",
        "AM" : "Armenia",
        "AO" : "Angola",
        "AQ" : "Antarctica",
        "AR" : "Argentina",
        "AS" : "American Samoa",
        "AT" : "Austria",
        "AU" : "Australia",
        "AW" : "Aruba",
        "AX" : "Åland Islands",
        "AZ" : "Azerbaijan",
        "BA" : "Bosnia and Herzegovina",
        "BB" : "Barbados",
        "BD" : "Bangladesh",
        "BE" : "Belgium",
        "BF" : "Burkina Faso",
        "BG" : "Bulgaria",
        "BH" : "Bahrain",
        "BI" : "Burundi",
        "BJ" : "Benin",
        "BL" : "Saint Barthélemy",
        "BM" : "Bermuda",
        "BN" : "Brunei Darussalam",
        "BO" : "Plurinational State of Bolivia",
        "BQ" : "Sint Eustatius and Saba Bonaire",
        "BR" : "Brazil",
        "BS" : "Bahamas",
        "BT" : "Bhutan",
        "BV" : "Bouvet Island",
        "BW" : "Botswana",
        "BY" : "Belarus",
        "BZ" : "Belize",
        "CA" : "Canada",
        "CC" : "Cocos (Keeling) Islands",
        "CD" : "The Democratic Republic of the Congo",
        "CF" : "Central African Republic",
        "CG" : "Congo",
        "CH" : "Switzerland",
        "CI" : "Côte d'Ivoire",
        "CK" : "Cook Islands",
        "CL" : "Chile",
        "CM" : "Cameroon",
        "CN" : "China",
        "CO" : "Colombia",
        "CR" : "Costa Rica",
        "CU" : "Cuba",
        "CV" : "Cape Verde",
        "CW" : "Curaçao",
        "CX" : "Christmas Island",
        "CY" : "Cyprus",
        "CZ" : "Czech Republic",
        "DE" : "Germany",
        "DJ" : "Djibouti",
        "DK" : "Denmark",
        "DM" : "Dominica",
        "DO" : "Dominican Republic",
        "DZ" : "Algeria",
        "EC" : "Ecuador",
        "EE" : "Estonia",
        "EG" : "Egypt",
        "EH" : "Western Sahara",
        "ER" : "Eritrea",
        "ES" : "Spain",
        "ET" : "Ethiopia",
        "FI" : "Finland",
        "FJ" : "Fiji",
        "FK" : "Falkland Islands (Malvinas)",
        "FM" : "Federated States of Micronesia",
        "FO" : "Faroe Islands",
        "FR" : "France",
        "GA" : "Gabon",
        "GB" : "United Kingdom",
        "GD" : "Grenada",
        "GE" : "Georgia",
        "GF" : "French Guiana",
        "GG" : "Guernsey",
        "GH" : "Ghana",
        "GI" : "Gibraltar",
        "GL" : "Greenland",
        "GM" : "Gambia",
        "GN" : "Guinea",
        "GP" : "Guadeloupe",
        "GQ" : "Equatorial Guinea",
        "GR" : "Greece",
        "GS" : "South Georgia and the South Sandwich Islands",
        "GT" : "Guatemala",
        "GU" : "Guam",
        "GW" : "Guinea-Bissau",
        "GY" : "Guyana",
        "HK" : "Hong Kong",
        "HM" : "Heard Island and McDonald Islands",
        "HN" : "Honduras",
        "HR" : "Croatia",
        "HT" : "Haiti",
        "HU" : "Hungary",
        "ID" : "Indonesia",
        "IE" : "Ireland",
        "IL" : "Israel",
        "IM" : "Isle of Man",
        "IN" : "India",
        "IO" : "British Indian Ocean Territory",
        "IQ" : "Iraq",
        "IR" : "Islamic Republic of Iran",
        "IS" : "Iceland",
        "IT" : "Italy",
        "JE" : "Jersey",
        "JM" : "Jamaica",
        "JO" : "Jordan",
        "JP" : "Japan",
        "KE" : "Kenya",
        "KG" : "Kyrgyzstan",
        "KH" : "Cambodia",
        "KI" : "Kiribati",
        "KM" : "Comoros",
        "KN" : "Saint Kitts and Nevis",
        "KP" : "Democratic People's Republic of Korea",
        "KR" : "Republic of Korea",
        "KW" : "Kuwait",
        "KY" : "Cayman Islands",
        "KZ" : "Kazakhstan",
        "LA" : "Lao People's Democratic Republic",
        "LB" : "Lebanon",
        "LC" : "Saint Lucia",
        "LI" : "Liechtenstein",
        "LK" : "Sri Lanka",
        "LR" : "Liberia",
        "LS" : "Lesotho",
        "LT" : "Lithuania",
        "LU" : "Luxembourg",
        "LV" : "Latvia",
        "LY" : "Libya",
        "MA" : "Morocco",
        "MC" : "Monaco",
        "MD" : "Republic of Moldova",
        "ME" : "Montenegro",
        "MF" : "Saint Martin (French part)",
        "MG" : "Madagascar",
        "MH" : "Marshall Islands",
        "MK" : "The Former Yugoslav Republic of Macedonia",
        "ML" : "Mali",
        "MM" : "Myanmar",
        "MN" : "Mongolia",
        "MO" : "Macao",
        "MP" : "Northern Mariana Islands",
        "MQ" : "Martinique",
        "MR" : "Mauritania",
        "MS" : "Montserrat",
        "MT" : "Malta",
        "MU" : "Mauritius",
        "MV" : "Maldives",
        "MW" : "Malawi",
        "MX" : "Mexico",
        "MY" : "Malaysia",
        "MZ" : "Mozambique",
        "NA" : "Namibia",
        "NC" : "New Caledonia",
        "NE" : "Niger",
        "NF" : "Norfolk Island",
        "NG" : "Nigeria",
        "NI" : "Nicaragua",
        "NL" : "Netherlands",
        "NO" : "Norway",
        "NP" : "Nepal",
        "NR" : "Nauru",
        "NU" : "Niue",
        "NZ" : "New Zealand",
        "OM" : "Oman",
        "PA" : "Panama",
        "PE" : "Peru",
        "PF" : "French Polynesia",
        "PG" : "Papua New Guinea",
        "PH" : "Philippines",
        "PK" : "Pakistan",
        "PL" : "Poland",
        "PM" : "Saint Pierre and Miquelon",
        "PN" : "Pitcairn",
        "PR" : "Puerto Rico",
        "PS" : "State of Palestine",
        "PT" : "Portugal",
        "PW" : "Palau",
        "PY" : "Paraguay",
        "QA" : "Qatar",
        "RE" : "Réunion",
        "RO" : "Romania",
        "RS" : "Serbia",
        "RU" : "Russian Federation",
        "RW" : "Rwanda",
        "SA" : "Saudi Arabia",
        "SB" : "Solomon Islands",
        "SC" : "Seychelles",
        "SD" : "Sudan",
        "SE" : "Sweden",
        "SG" : "Singapore",
        "SH" : "Saint Helena, Ascension and Tristan da Cunha",
        "SI" : "Slovenia",
        "SJ" : "Svalbard and Jan Mayen",
        "SK" : "Slovakia",
        "SL" : "Sierra Leone",
        "SM" : "San Marino",
        "SN" : "Senegal",
        "SO" : "Somalia",
        "SR" : "Suriname",
        "SS" : "South Sudan",
        "ST" : "Sao Tome and Principe",
        "SV" : "El Salvador",
        "SX" : "Sint Maarten",
        "SY" : "Syrian Arab Republic",
        "SZ" : "Swaziland",
        "TC" : "Turks and Caicos Islands",
        "TD" : "Chad",
        "TF" : "French Southern Territories",
        "TG" : "Togo",
        "TH" : "Thailand",
        "TJ" : "Tajikistan",
        "TK" : "Tokelau",
        "TL" : "Timor-Leste",
        "TM" : "Turkmenistan",
        "TN" : "Tunisia",
        "TO" : "Tonga",
        "TR" : "Turkey",
        "TT" : "Trinidad and Tobago",
        "TV" : "Tuvalu",
        "TW" : "Taiwan, Province of China",
        "TZ" : "United Republic of Tanzania",
        "UA" : "Ukraine",
        "UG" : "Uganda",
        "UM" : "United States Minor Outlying Islands",
        "US" : "United States",
        "UY" : "Uruguay",
        "UZ" : "Uzbekistan",
        "VA" : "Holy See (Vatican City State)",
        "VC" : "Saint Vincent and the Grenadines",
        "VE" : "Bolivarian Republic of Venezuela",
        "VG" : "British Virgin Islands",
        "VI" : "U.S. Virgin Islands",
        "VN" : "Viet Nam",
        "VU" : "Vanuatu",
        "WF" : "Wallis and Futuna",
        "WS" : "Samoa",
        "YE" : "Yemen",
        "YT" : "Mayotte",
        "ZA" : "South Africa",
        "ZM" : "Zambia",
        "ZW" : "Zimbabwe",
    }

# END Class isoCountry



# -*- coding: utf-8 -*-

from nw.gui.about import GuiAbout
from nw.gui.build import GuiBuildNovel
from nw.gui.doceditor import GuiDocEditor
from nw.gui.docmerge import GuiDocMerge
from nw.gui.docsplit import GuiDocSplit
from nw.gui.docviewer import GuiDocViewer, GuiDocViewDetails
from nw.gui.itemdetails import GuiItemDetails
from nw.gui.itemeditor import GuiItemEditor
from nw.gui.mainmenu import GuiMainMenu
from nw.gui.outline import GuiOutline
from nw.gui.outlinedetails import GuiOutlineDetails
from nw.gui.preferences import GuiPreferences
from nw.gui.projload import GuiProjectLoad
from nw.gui.projsettings import GuiProjectSettings
from nw.gui.projtree import GuiProjectTree
from nw.gui.projwizard import GuiProjectWizard
from nw.gui.statusbar import GuiMainStatus
from nw.gui.theme import GuiTheme
from nw.gui.writingstats import GuiWritingStats

__all__ = [
    "GuiAbout",
    "GuiBuildNovel",
    "GuiDocEditor",
    "GuiDocMerge",
    "GuiDocSplit",
    "GuiDocViewDetails",
    "GuiDocViewer",
    "GuiItemDetails",
    "GuiItemEditor",
    "GuiMainMenu",
    "GuiMainStatus",
    "GuiOutline",
    "GuiOutlineDetails",
    "GuiPreferences",
    "GuiProjectLoad",
    "GuiProjectSettings",
    "GuiProjectTree",
    "GuiProjectWizard",
    "GuiTheme",
    "GuiWritingStats",
]



# -*- coding: utf-8 -*-
"""novelWriter GUI Main Window StatusBar

 novelWriter – GUI Main Window StatusBar
=========================================
 Class holding the main window status bar

 File History:
 Created: 2019-04-20 [0.0.1]

 This file is a part of novelWriter
 Copyright 2018–2020, Veronica Berglyd Olsen

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program. If not, see <https://www.gnu.org/licenses/>.
"""

import nw
import logging

from time import time

from PyQt5.QtCore import QTimer
from PyQt5.QtGui import QColor, QPainter
from PyQt5.QtWidgets import qApp, QStatusBar, QLabel, QAbstractButton

from nw.core import NWSpellCheck
from nw.common import formatTime

logger = logging.getLogger(__name__)

class GuiMainStatus(QStatusBar):

    def __init__(self, theParent):
        QStatusBar.__init__(self, theParent)

        logger.debug("Initialising GuiMainStatus ...")

        self.mainConf  = nw.CONFIG
        self.theParent = theParent
        self.theTheme  = theParent.theTheme
        self.refTime   = None

        self.projWords = 0
        self.sessWords = 0

        colNone  = QColor(*self.theTheme.statNone)
        colTrue  = QColor(*self.theTheme.statUnsaved)
        colFalse = QColor(*self.theTheme.statSaved)

        iPx = self.theTheme.baseIconSize

        # Permanent Widgets
        # =================

        xM = self.mainConf.pxInt(8)

        ## The Spell Checker Language
        self.langIcon = QLabel("")
        self.langText = QLabel("None")
        self.langIcon.setPixmap(self.theTheme.getPixmap("status_lang", (iPx, iPx)))
        self.langIcon.setContentsMargins(0, 0, 0, 0)
        self.langText.setContentsMargins(0, 0, xM, 0)
        self.addPermanentWidget(self.langIcon)
        self.addPermanentWidget(self.langText)

        ## The Editor Status
        self.docIcon = StatusLED(colNone, colTrue, colFalse, iPx, iPx, self)
        self.docText = QLabel("Editor")
        self.docIcon.setContentsMargins(0, 0, 0, 0)
        self.docText.setContentsMargins(0, 0, xM, 0)
        self.addPermanentWidget(self.docIcon)
        self.addPermanentWidget(self.docText)

        ## The Project Status
        self.projIcon = StatusLED(colNone, colTrue, colFalse, iPx, iPx, self)
        self.projText = QLabel("Project")
        self.projIcon.setContentsMargins(0, 0, 0, 0)
        self.projText.setContentsMargins(0, 0, xM, 0)
        self.addPermanentWidget(self.projIcon)
        self.addPermanentWidget(self.projText)

        ## The Project and Session Stats
        self.statsIcon = QLabel()
        self.statsText = QLabel("")
        self.statsIcon.setPixmap(self.theTheme.getPixmap("status_stats", (iPx, iPx)))
        self.statsIcon.setContentsMargins(0, 0, 0, 0)
        self.statsText.setContentsMargins(0, 0, xM, 0)
        self.addPermanentWidget(self.statsIcon)
        self.addPermanentWidget(self.statsText)

        ## The Session Clock
        ### Set the mimimum width so the label doesn't rescale every second
        self.timeIcon = QLabel()
        self.timeText = QLabel("")
        self.timeIcon.setPixmap(self.theTheme.getPixmap("status_time", (iPx, iPx)))
        self.timeText.setToolTip("Session Time")
        self.timeText.setMinimumWidth(self.theTheme.getTextWidth("00:00:00:"))
        self.timeIcon.setContentsMargins(0, 0, 0, 0)
        self.timeText.setContentsMargins(0, 0, 0, 0)
        self.addPermanentWidget(self.timeIcon)
        self.addPermanentWidget(self.timeText)

        # Other Settings
        self.setSizeGripEnabled(True)

        # Start the Clock
        self.sessionTimer = QTimer()
        self.sessionTimer.setInterval(1000)
        self.sessionTimer.timeout.connect(self._updateTime)
        self.sessionTimer.start()

        logger.debug("GuiMainStatus initialisation complete")

        self.clearStatus()

        return

    def clearStatus(self):
        """Reset all widgets on the status bar to default values.
        """
        self.setRefTime(None)
        self.setStats(0, 0)
        self.setProjectStatus(None)
        self.setDocumentStatus(None)
        self._updateTime()
        return True

    ##
    #  Setters
    ##

    def setRefTime(self, theTime):
        """Set the reference time for the status bar clock.
        """
        self.refTime = theTime
        return

    def setStatus(self, theMessage, timeOut=20.0):
        """Set the status bar message to display for 'timeOut' seconds.
        """
        self.showMessage(theMessage, int(timeOut*1000))
        qApp.processEvents()
        return

    def setLanguage(self, theLanguage, theProvider=""):
        """Set the language code for the spell checker.
        """
        if theLanguage is None:
            self.langText.setText("None")
            self.langText.setToolTip("")
        else:
            self.langText.setText(NWSpellCheck.expandLanguage(theLanguage))
            self.langText.setToolTip(
                "Provider: %s" % (theProvider if theProvider else "unknown")
            )
        return

    def setProjectStatus(self, isChanged):
        """Set the project status colour icon.
        """
        self.projIcon.setState(isChanged)
        return

    def setDocumentStatus(self, isChanged):
        """Set the document status colour icon.
        """
        self.docIcon.setState(isChanged)
        return

    def setStats(self, pWC, sWC):
        """Set the current project statistics.
        """
        self.projWords = pWC
        self.sessWords = sWC
        self._updateStats()
        return

    ##
    #  Internal Functions
    ##

    def _updateStats(self):
        """Update statistics.
        """
        self.statsText.setToolTip(
            "Project word count (session change)"
        )
        self.statsText.setText(
            f"Words: {self.projWords:n} ({self.sessWords:+n})"
        )
        return

    def _updateTime(self):
        """Update the session clock.
        """
        if self.refTime is None:
            self.timeText.setText("00:00:00")
        else:
            self.timeText.setText(formatTime(round(time() - self.refTime)))
        return

# END Class GuiMainStatus

class StatusLED(QAbstractButton):

    def __init__(self, colNone, colTrue, colFalse, sW, sH, parent=None):
        super().__init__(parent=parent)

        self.colNone  = colNone
        self.colTrue  = colTrue
        self.colFalse = colFalse
        self._theCol  = colNone

        self.setFixedWidth(sW)
        self.setFixedHeight(sH)

        return

    ##
    #  Setters
    ##

    def setState(self, theState):
        """Set the colour state.
        """
        if theState is None:
            self._theCol = self.colNone
        elif theState:
            self._theCol = self.colTrue
        elif not theState:
            self._theCol = self.colFalse
        else:
            self._theCol = self.colNone
        self.update()
        return

    ##
    #  Events
    ##

    def paintEvent(self, event):
        """Drawing the LED.
        """
        qPalette = self.palette()
        qPaint = QPainter(self)
        qPaint.setRenderHint(QPainter.Antialiasing, True)
        qPaint.setPen(qPalette.dark().color())
        qPaint.setBrush(self._theCol)
        qPaint.setOpacity(1.0)
        qPaint.drawEllipse(1, 1, self.width() - 2, self.height() - 2)
        return

# END Class StatusLED



# -*- coding: utf-8 -*-
"""novelWriter GUI Document Editor

 novelWriter – GUI Document Editor
===================================
 Class holding the main document editor

 File History:
 Created:   2018-09-29 [0.0.1]  GuiDocEditor
 Created:   2019-04-22 [0.0.1]  BackgroundWordCounter
 Created:   2019-09-29 [0.2.1]  GuiDocEditSearch
 Created:   2020-04-25 [0.4.5]  GuiDocEditHeader
 Rewritten: 2020-06-15 [0.9.0]  GuiDocEditSearch
 Created:   2020-06-27 [0.10.0] GuiDocEditFooter
 Rewritten: 2020-10-07 [1.0b3]  BackgroundWordCounter

 This file is a part of novelWriter
 Copyright 2018–2020, Veronica Berglyd Olsen

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program. If not, see <https://www.gnu.org/licenses/>.
"""

import nw
import logging

from time import time

from PyQt5.QtCore import (
    Qt, QSize, QTimer, pyqtSlot, pyqtSignal, QRegExp, QRegularExpression,
    QPointF, QObject, QRunnable, QPropertyAnimation
)
from PyQt5.QtGui import (
    QTextCursor, QTextOption, QKeySequence, QFont, QColor, QPalette,
    QTextDocument, QCursor, QPixmap
)
from PyQt5.QtWidgets import (
    qApp, QTextEdit, QAction, QMenu, QShortcut, QMessageBox, QWidget, QLabel,
    QToolBar, QToolButton, QHBoxLayout, QGridLayout, QLineEdit, QPushButton,
    QFrame
)

from nw.core import NWDoc, NWSpellSimple, countWords
from nw.gui.dochighlight import GuiDocHighlighter
from nw.common import transferCase
from nw.constants import (
    nwConst, nwAlert, nwUnicode, nwDocAction, nwDocInsert, nwItemClass
)

logger = logging.getLogger(__name__)

class GuiDocEditor(QTextEdit):

    MOVE_KEYS = (
        Qt.Key_Left, Qt.Key_Right, Qt.Key_Up, Qt.Key_Down,
        Qt.Key_PageUp, Qt.Key_PageDown
    )

    def __init__(self, theParent):
        QTextEdit.__init__(self, theParent)

        logger.debug("Initialising GuiDocEditor ...")

        # Class Variables
        self.mainConf   = nw.CONFIG
        self.theParent  = theParent
        self.theTheme   = theParent.theTheme
        self.theProject = theParent.theProject
        self.nwDocument = NWDoc(self.theProject, self.theParent)

        self.docChanged = False # Flag for changed status of document
        self.spellCheck = False # Flag for spell checking enabled
        self.theHandle  = None  # The handle of the open file
        self.theDict    = None  # The current spell check dictionary
        self.nonWord    = "\"'" # Characters to not include in spell checking

        # Document Variables
        self.charCount  = 0     # Character count
        self.wordCount  = 0     # Word count
        self.paraCount  = 0     # Paragraph count
        self.lastEdit   = 0     # Time stamp of last edit
        self.lastFind   = None  # Position of the last found search word
        self.bigDoc     = False # Flag for very large document size
        self.doReplace  = False # Switch to temporarily disable auto-replace
        self.queuePos   = None  # Used for delayed change of cursor position

        # Typography
        self.typDQOpen  = self.mainConf.fmtDoubleQuotes[0]
        self.typDQClose = self.mainConf.fmtDoubleQuotes[1]
        self.typSQOpen  = self.mainConf.fmtSingleQuotes[0]
        self.typSQClose = self.mainConf.fmtSingleQuotes[1]

        # Core Elements and Signals
        self.qDocument = self.document()
        self.qDocument.contentsChange.connect(self._docChange)
        self.qDocument.documentLayout().documentSizeChanged.connect(self._docSizeChanged)

        # Document Title
        self.docHeader = GuiDocEditHeader(self)
        self.docFooter = GuiDocEditFooter(self)
        self.docSearch = GuiDocEditSearch(self)

        # Syntax
        self.hLight = GuiDocHighlighter(self.qDocument, self.theParent)

        # Context Menu
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self._openContextMenu)

        # Editor Settings
        self.setMinimumWidth(self.mainConf.pxInt(300))
        self.setAutoFillBackground(True)
        self.setAcceptRichText(False)

        # Custom Shortcuts
        QShortcut(
            QKeySequence("Ctrl+."),
            self,
            context=Qt.WidgetShortcut,
            activated=self._openSpellContext
        )
        QShortcut(
            Qt.Key_Return | Qt.ControlModifier,
            self,
            context=Qt.WidgetShortcut,
            activated=self._followTag
        )
        QShortcut(
            Qt.Key_Enter | Qt.ControlModifier,
            self,
            context=Qt.WidgetShortcut,
            activated=self._followTag
        )

        # Set Up Word Counter
        self.wcInterval = self.mainConf.wordCountTimer
        self.wcTimer = QTimer()
        self.wcTimer.setInterval(int(self.wcInterval*1000))
        self.wcTimer.timeout.connect(self._runCounter)

        self.wCounter = BackgroundWordCounter(self)
        self.wCounter.setAutoDelete(False)
        self.wCounter.signals.countsReady.connect(self._updateCounts)

        self.initEditor()

        logger.debug("GuiDocEditor initialisation complete")

        return

    def clearEditor(self):
        """Clear the current document and reset all document related
        flags and counters.
        """
        self.nwDocument.clearDocument()
        self.setReadOnly(True)
        self.clear()
        self.wcTimer.stop()

        self.theHandle = None
        self.charCount = 0
        self.wordCount = 0
        self.paraCount = 0
        self.lastEdit  = 0
        self.lastFind  = None
        self.bigDoc    = False
        self.doReplace = False
        self.queuePos  = None

        self.setDocumentChanged(False)
        self.docHeader.setTitleFromHandle(self.theHandle)
        self.docFooter.setHandle(self.theHandle)

        return True

    def initEditor(self):
        """Initialise or re-initialise the editor with the user's
        settings. This function is both called when the editor is
        created, and when the user changes the main editor preferences.
        """
        # Some Constants
        self.nonWord  = "\"'"
        self.nonWord += "".join(self.mainConf.fmtDoubleQuotes)
        self.nonWord += "".join(self.mainConf.fmtSingleQuotes)

        # Reload spell check and dictionaries
        self._setupSpellChecking()
        self.setDictionaries()

        # Set font
        theFont = QFont()
        if self.mainConf.textFont is None:
            # If none is defined, set the default back to config
            self.mainConf.textFont = self.qDocument.defaultFont().family()
        theFont.setFamily(self.mainConf.textFont)
        theFont.setPointSize(self.mainConf.textSize)
        self.setFont(theFont)

        # Set the widget colours to match syntax theme
        mainPalette = self.palette()
        mainPalette.setColor(QPalette.Window, QColor(*self.theTheme.colBack))
        mainPalette.setColor(QPalette.Base, QColor(*self.theTheme.colBack))
        mainPalette.setColor(QPalette.Text, QColor(*self.theTheme.colText))
        self.setPalette(mainPalette)

        docPalette = self.viewport().palette()
        docPalette.setColor(QPalette.Base, QColor(*self.theTheme.colBack))
        docPalette.setColor(QPalette.Text, QColor(*self.theTheme.colText))
        self.viewport().setPalette(docPalette)

        self.docHeader.matchColours()
        self.docFooter.matchColours()

        # Set default text margins
        cM = self.mainConf.getTextMargin()
        self.qDocument.setDocumentMargin(0)
        self.setViewportMargins(cM, cM, cM, cM)

        # Also set the document text options for the document text flow
        theOpt = QTextOption()

        if self.mainConf.doJustify:
            theOpt.setAlignment(Qt.AlignJustify)
        if self.mainConf.showTabsNSpaces:
            theOpt.setFlags(theOpt.flags() | QTextOption.ShowTabsAndSpaces)
        if self.mainConf.showLineEndings:
            theOpt.setFlags(theOpt.flags() | QTextOption.ShowLineAndParagraphSeparators)

        self.qDocument.setDefaultTextOption(theOpt)

        # Scroll bars
        if self.mainConf.hideVScroll:
            self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        else:
            self.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)

        if self.mainConf.hideHScroll:
            self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        else:
            self.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)

        # Refresh the tab stops
        if self.mainConf.verQtValue >= 51000:
            self.setTabStopDistance(self.mainConf.getTabWidth())
        else:
            self.setTabStopWidth(self.mainConf.getTabWidth())

        # Initialise the syntax highlighter
        self.hLight.initHighlighter()

        # If we have a document open, we should reload it in case the
        # font changed, otherwise we just clear the editor entirely,
        # which makes it read only.
        if self.theHandle is not None:
            self.redrawText()
        else:
            self.clearEditor()

        return True

    def loadText(self, tHandle, tLine=None, showStatus=True):
        """Load text from a document into the editor. If we have an io
        error, we must handle this and clear the editor so that we don't
        risk overwriting the file if it exists. This can for instance
        happen of the file contains binary elements or an encoding that
        novelWriter does not support. If load is successful, or the
        document is new (empty string), we set up the editor for editing
        the file.
        """
        theDoc = self.nwDocument.openDocument(tHandle, showStatus=showStatus)
        if theDoc is None:
            # There was an io error
            self.clearEditor()
            return False

        docSize = len(theDoc)
        if docSize > nwConst.MAX_DOCSIZE:
            self.theParent.makeAlert((
                "The document you are trying to open is too big. "
                "The document size is %.2f\u202fMB. "
                "The maximum size allowed is %.2f\u202fMB."
            ) % (docSize/1.0e6, nwConst.MAX_DOCSIZE/1.0e6), nwAlert.ERROR)
            self.clearEditor()
            return False

        qApp.setOverrideCursor(QCursor(Qt.WaitCursor))
        self.hLight.setHandle(tHandle)

        # Check that the document is not too big for full, initial spell
        # checking. If it is too big, we switch to only check as we type
        self._checkDocSize(docSize)
        spTemp = self.hLight.spellCheck
        if self.bigDoc:
            self.hLight.spellCheck = False

        bfTime = time()
        self._allowAutoReplace(False)
        self.setPlainText(theDoc)
        qApp.processEvents()

        self._allowAutoReplace(True)
        afTime = time()
        logger.debug("Document highlighted in %.3f ms" % (1000*(afTime-bfTime)))

        self.lastEdit = time()
        self._runCounter()
        self.wcTimer.start()
        self.setDocumentChanged(False)
        self.theHandle = tHandle

        self.setReadOnly(False)
        self.docHeader.setTitleFromHandle(self.theHandle)
        self.docFooter.setHandle(self.theHandle)
        self.updateDocMargins()
        self.hLight.spellCheck = spTemp

        theItem = self.nwDocument.getCurrentItem()
        if tLine is None and theItem is not None:
            # For large documents we queue the repositioning until the
            # document layout has grown past the point we want to move
            # the cursor to. This makes the loading significantly
            # faster.
            if docSize > 50000:
                self.queuePos = theItem.cursorPos
            else:
                self.setCursorPosition(theItem.cursorPos)
        else:
            self.setCursorLine(tLine)

        self.docFooter.updateLineCount()
        self.lengthLast = self.qDocument.characterCount()

        qApp.restoreOverrideCursor()

        return True

    def updateTagHighLighting(self, forceBigDoc=False):
        """Rerun the syntax highlighter on all meta data lines.
        """
        self.hLight.rehighlightByType(GuiDocHighlighter.BLOCK_META)
        return

    def redrawText(self):
        """Redraw the text by marking the document content as "dirty".
        """
        self.qDocument.markContentsDirty(0, self.qDocument.characterCount())
        return

    def replaceText(self, theText):
        """Replaces the text of the current document with the provided
        text. This also clears undo history.
        """
        docSize = len(theText)
        if docSize > nwConst.MAX_DOCSIZE:
            self.theParent.makeAlert((
                "The text you are trying to add is too big. "
                "The text size is %.2f\u202fMB. "
                "The maximum size allowed is %.2f\u202fMB."
            ) % (docSize/1.0e6, nwConst.MAX_DOCSIZE/1.0e6), nwAlert.ERROR)
            return False

        qApp.setOverrideCursor(QCursor(Qt.WaitCursor))
        self.setPlainText(theText)
        self.setDocumentChanged(True)
        self.updateDocMargins()
        qApp.restoreOverrideCursor()

        return True

    def saveText(self):
        """Save the text currently in the editor to the NWDoc object,
        and update the NWItem meta data.
        """
        theItem = self.nwDocument.getCurrentItem()
        if theItem is None:
            return False

        docText = self.getText()

        cC, wC, pC = countWords(docText)
        self._updateCounts(cC, wC, pC)

        theItem.setCharCount(self.charCount)
        theItem.setWordCount(self.wordCount)
        theItem.setParaCount(self.paraCount)

        self.saveCursorPosition()
        self.nwDocument.saveDocument(docText)
        self.setDocumentChanged(False)

        self.theParent.theIndex.scanText(theItem.itemHandle, docText)

        return True

    def updateDocMargins(self):
        """Automatically adjust the margins so the text is centred if
        Config.textFixedW is enabled or we're in Focus Mode. Otherwise,
        just ensure the margins are set correctly.
        """
        wW = self.width()
        wH = self.height()
        cM = self.mainConf.getTextMargin()

        vBar = self.verticalScrollBar()
        sW = vBar.width() if vBar.isVisible() else 0

        hBar = self.horizontalScrollBar()
        sH = hBar.height() if hBar.isVisible() else 0

        if self.mainConf.textFixedW or self.theParent.isFocusMode:
            if self.theParent.isFocusMode:
                tW = self.mainConf.getFocusWidth()
            else:
                tW = self.mainConf.getTextWidth()
            tM = (wW - sW - tW)//2
            if tM < cM:
                tM = cM
        else:
            tM = cM

        tB = self.frameWidth()
        tW = wW - 2*tB - sW
        tH = self.docHeader.height()
        fH = self.docFooter.height()
        fY = wH - fH - tB - sH
        self.docHeader.setGeometry(tB, tB, tW, tH)
        self.docFooter.setGeometry(tB, fY, tW, fH)

        if self.docSearch.isVisible():
            rH = self.docSearch.height()
            rW = self.docSearch.width()
            rL = wW - sW - rW - 2*tB
            self.docSearch.move(rL, 2*tB)
        else:
            rH = 0

        uM = max(cM, tH, rH)
        lM = max(cM, fH)
        self.setViewportMargins(tM, uM, tM, lM)

        if self.mainConf.scrollPastEnd:
            docFrame = self.qDocument.rootFrame().frameFormat()
            docFrame.setBottomMargin(max(0, 0.6*(wH - uM - lM - 4*tB)))
            self.qDocument.rootFrame().setFrameFormat(docFrame)

        return

    def updateDocInfo(self, tHandle):
        """Called when an item label is changed to check if the document
        title bar needs updating,
        """
        if tHandle == self.theHandle:
            self.docHeader.setTitleFromHandle(self.theHandle)
            self.docFooter.updateInfo()
            self.updateDocMargins()
        return

    ##
    #  Setters and Getters
    ##

    def setDocumentChanged(self, bValue):
        """Keeps track of the document changed variable, and ensures
        that the corresponding icon on the status bar shows the same
        status.
        """
        self.docChanged = bValue
        self.theParent.statusBar.setDocumentStatus(self.docChanged)
        return self.docChanged

    def getText(self):
        """Get the text content of the current document. This method
        uses QTextEdit->toPlainText for Qt versions lower than 5.9, and
        the QDocument->toRawText for higher version. The latter
        preserves non-breaking spaces, which the former does not.
        We still want to get rid of page and line separators though.
        See: https://doc.qt.io/qt-5/qtextdocument.html#toPlainText
        """
        if self.mainConf.verQtValue >= 50900:
            theText = self.qDocument.toRawText()
            theText = theText.replace(nwUnicode.U_LSEP, "\n") # Line separators
            theText = theText.replace(nwUnicode.U_PSEP, "\n") # Paragraph separators
        else:
            theText = self.toPlainText()
        return theText

    def setCursorPosition(self, thePosition):
        """Move the cursor to a given position in the document.
        """
        if not isinstance(thePosition, int):
            return False

        if thePosition >= 0:
            theCursor = self.textCursor()
            theCursor.setPosition(thePosition)
            self.setTextCursor(theCursor)
            self.docFooter.updateLineCount()

        return True

    def getCursorPosition(self):
        """Find the cursor position in the document. If the editor has a
        selection, return the position of the end of the selection.
        """
        return self.textCursor().selectionEnd()

    def saveCursorPosition(self):
        """Save the cursor position to the current project item object.
        """
        theItem = self.nwDocument.getCurrentItem()
        if theItem is not None:
            cursPos = self.getCursorPosition()
            theItem.setCursorPos(cursPos)
        return

    def setCursorLine(self, theLine):
        """Move the cursor to a given line in the document.
        """
        if not isinstance(theLine, int):
            return False
        if theLine >= 0:
            theBlock = self.qDocument.findBlockByLineNumber(theLine)
            if theBlock:
                self.setCursorPosition(theBlock.position())
                self.docFooter.updateLineCount()
                logger.verbose("Cursor moved to line %d" % theLine)
        return True

    ##
    #  Spell Checking
    ##

    def setDictionaries(self):
        """Set the spell checker dictionary language, and update the
        status bar to show the one actually loaded by the spell checker
        class.
        """
        if self.theProject.projLang is None:
            theLang = self.mainConf.spellLanguage
        else:
            theLang = self.theProject.projLang

        self.theDict.setLanguage(theLang, self.theProject.projDict)

        aLang, aName = self.theDict.describeDict()
        self.theParent.statusBar.setLanguage(
            aLang, "%s [%s]" % (self.mainConf.spellTool.title(), aName.title())
        )

        if not self.bigDoc:
            self.spellCheckDocument()

        return True

    def setSpellCheck(self, theMode):
        """This is the master spell check setting function, and this one
        should call all other setSpellCheck functions in other classes.
        If the spell check mode (theMode) is not defined (None), then
        toggle the current status saved in this class.
        """
        if theMode is None:
            theMode = not self.spellCheck

        if self.theDict.spellLanguage is None:
            theMode = False

        self.spellCheck = theMode
        self.theParent.mainMenu.setSpellCheck(theMode)
        self.theProject.setSpellCheck(theMode)
        self.hLight.setSpellCheck(theMode)
        if not self.bigDoc:
            self.spellCheckDocument()

        logger.verbose("Spell check is set to %s" % str(theMode))

        return True

    def spellCheckDocument(self):
        """Rerun the highlighter to update spell checking status of the
        currently loaded text. The fastest way to do this, at least as
        of Qt 5.13, is to clear the text and put it back.
        """
        logger.verbose("Running spell checker")
        if self.spellCheck:
            bfTime = time()
            qApp.setOverrideCursor(QCursor(Qt.WaitCursor))
            if self.bigDoc:
                theText = self.getText()
                self.setPlainText(theText)
            else:
                self.hLight.rehighlight()
            qApp.restoreOverrideCursor()
            afTime = time()
            logger.debug(
                "Document highlighted in %.3f ms" % (1000*(afTime-bfTime))
            )
            self.theParent.statusBar.showMessage("Spell check complete")

        return True

    ##
    #  General Class Methods
    ##

    def docAction(self, theAction):
        """Perform an action on the current document based on an action
        flag. This is just a single entry point wrapper function to
        ensure all the feature functions get the correct information
        passed to it without having to consider the internal logic of
        this class when calling these actions from other classes.
        """
        logger.verbose("Requesting action: %s" % theAction.name)
        if self.theHandle is None:
            logger.error("No document open")
            return False

        self._allowAutoReplace(False)
        if theAction == nwDocAction.UNDO:
            self.undo()
        elif theAction == nwDocAction.REDO:
            self.redo()
        elif theAction == nwDocAction.CUT:
            self.cut()
        elif theAction == nwDocAction.COPY:
            self.copy()
        elif theAction == nwDocAction.PASTE:
            self.paste()
        elif theAction == nwDocAction.EMPH:
            self._toggleFormat(1, "_")
        elif theAction == nwDocAction.STRONG:
            self._toggleFormat(2, "*")
        elif theAction == nwDocAction.STRIKE:
            self._toggleFormat(2, "~")
        elif theAction == nwDocAction.S_QUOTE:
            self._wrapSelection(self.typSQOpen, self.typSQClose)
        elif theAction == nwDocAction.D_QUOTE:
            self._wrapSelection(self.typDQOpen, self.typDQClose)
        elif theAction == nwDocAction.SEL_ALL:
            self._makeSelection(QTextCursor.Document)
        elif theAction == nwDocAction.SEL_PARA:
            self._makeSelection(QTextCursor.BlockUnderCursor)
        elif theAction == nwDocAction.FIND:
            self._beginSearch()
        elif theAction == nwDocAction.REPLACE:
            self._beginReplace()
        elif theAction == nwDocAction.GO_NEXT:
            self._findNext()
        elif theAction == nwDocAction.GO_PREV:
            self._findNext(isBackward=True)
        elif theAction == nwDocAction.REPL_NEXT:
            self._replaceNext()
        elif theAction == nwDocAction.BLOCK_H1:
            self._formatBlock(nwDocAction.BLOCK_H1)
        elif theAction == nwDocAction.BLOCK_H2:
            self._formatBlock(nwDocAction.BLOCK_H2)
        elif theAction == nwDocAction.BLOCK_H3:
            self._formatBlock(nwDocAction.BLOCK_H3)
        elif theAction == nwDocAction.BLOCK_H4:
            self._formatBlock(nwDocAction.BLOCK_H4)
        elif theAction == nwDocAction.BLOCK_COM:
            self._formatBlock(nwDocAction.BLOCK_COM)
        elif theAction == nwDocAction.BLOCK_TXT:
            self._formatBlock(nwDocAction.BLOCK_TXT)
        elif theAction == nwDocAction.REPL_SNG:
            self._replaceQuotes("'", self.typSQOpen, self.typSQClose)
        elif theAction == nwDocAction.REPL_DBL:
            self._replaceQuotes("\"", self.typDQOpen, self.typDQClose)
        else:
            logger.debug("Unknown or unsupported document action %s" % str(theAction))
            self._allowAutoReplace(True)
            return False

        self._allowAutoReplace(True)

        return True

    def isEmpty(self):
        """Wrapper function to check if the current document is empty.
        """
        return self.qDocument.isEmpty()

    def revealLocation(self):
        """Tell the user where on the file system the file in the editor
        is saved.
        """
        if self.theHandle is None:
            logger.error("No document open")
            return False

        msgBox = QMessageBox()
        msgBox.information(self, "File Location", (
            "File details for the currently open file<br>"
            "Handle: {handle:s}<br>"
            "Location: {fileLoc:s}"
        ).format(
            handle  = self.theHandle,
            fileLoc = str(self.nwDocument.getFileLocation())
        ))

        return

    def insertText(self, theInsert):
        """Insert a specific type of text at the cursor position.
        """
        if self.theHandle is None:
            logger.error("No document open")
            return False

        if isinstance(theInsert, str):
            theText = theInsert
        elif isinstance(theInsert, nwDocInsert):
            if theInsert == nwDocInsert.HARD_BREAK:
                theText = "  \n"
            elif theInsert == nwDocInsert.NB_SPACE:
                theText = nwUnicode.U_NBSP
            elif theInsert == nwDocInsert.THIN_SPACE:
                theText = nwUnicode.U_THNSP
            elif theInsert == nwDocInsert.THIN_NB_SPACE:
                theText = nwUnicode.U_THNBSP
            elif theInsert == nwDocInsert.SHORT_DASH:
                theText = nwUnicode.U_ENDASH
            elif theInsert == nwDocInsert.LONG_DASH:
                theText = nwUnicode.U_EMDASH
            elif theInsert == nwDocInsert.ELLIPSIS:
                theText = nwUnicode.U_HELLIP
            elif theInsert == nwDocInsert.MODAPOS_S:
                theText = nwUnicode.U_MAPOSS
            elif theInsert == nwDocInsert.QUOTE_LS:
                theText = self.typSQOpen
            elif theInsert == nwDocInsert.QUOTE_RS:
                theText = self.typSQClose
            elif theInsert == nwDocInsert.QUOTE_LD:
                theText = self.typDQOpen
            elif theInsert == nwDocInsert.QUOTE_RD:
                theText = self.typDQClose
            else:
                return False
        else:
            return False

        theCursor = self.textCursor()
        theCursor.beginEditBlock()
        theCursor.insertText(theText)
        theCursor.endEditBlock()

        return True

    def closeSearch(self):
        """Close the search box.
        """
        self.docSearch.closeSearch()
        return self.docSearch.isVisible()

    def toggleSearch(self):
        """Toggle the visibility of the search box.
        """
        if self.docSearch.isVisible():
            self.docSearch.closeSearch()
        else:
            self._beginSearch()
        return

    ##
    #  Document Events and Maintenance
    ##

    def keyPressEvent(self, keyEvent):
        """Intercept key press events.
        We need to intercept a few key sequences:
          * The return and enter key redirects here even if the search
            box has focus. Since we need these keys to continue search,
            we block any further interaction here while it's in focus.
          * The undo/redo/select all sequences bypasses the docAction
            pathway from the menu, so we redirect them back from here.
        """
        isReturn  = keyEvent.key() == Qt.Key_Return
        isReturn |= keyEvent.key() == Qt.Key_Enter
        if isReturn and self.docSearch.anyFocus():
            return
        elif keyEvent == QKeySequence.Redo:
            self.docAction(nwDocAction.REDO)
            return
        elif keyEvent == QKeySequence.Undo:
            self.docAction(nwDocAction.UNDO)
            return
        elif keyEvent == QKeySequence.SelectAll:
            self.docAction(nwDocAction.SEL_ALL)
            return

        if self.mainConf.autoScroll:

            cOld = self.cursorRect().center().y()
            QTextEdit.keyPressEvent(self, keyEvent)

            kMod = keyEvent.modifiers()
            okMod = kMod == Qt.NoModifier or kMod == Qt.ShiftModifier
            okKey = keyEvent.key() not in self.MOVE_KEYS
            if okMod and okKey:
                cNew = self.cursorRect().center().y()
                cMov = cNew - cOld
                mPos = self.mainConf.autoScrollPos * self.viewport().height() * 0.01
                if abs(cMov) > 0 and cOld > mPos:
                    # Move the scroll bar
                    vBar = self.verticalScrollBar()
                    doAnim = QPropertyAnimation(vBar, b"value", self)
                    doAnim.setDuration(120)
                    doAnim.setStartValue(vBar.value())
                    doAnim.setEndValue(vBar.value() + cMov)
                    doAnim.start()

        else:
            QTextEdit.keyPressEvent(self, keyEvent)

        self.docFooter.updateLineCount()

        return

    def focusNextPrevChild(self, toNext):
        """Capture the focus request from the tab key on the text
        editor. If the editor has focus, we do not change focus and
        allow the editor to insert a tab. If the search bar has focus,
        we forward the call to the search object.
        """
        if self.hasFocus():
            return False
        elif self.docSearch.isVisible():
            return self.docSearch.cycleFocus(toNext)
        return True

    def mouseReleaseEvent(self, theEvent):
        """If the mouse button is released and the control key is
        pressed, check if we're clicking on a tag, and trigger the
        follow tag function.
        """
        if qApp.keyboardModifiers() == Qt.ControlModifier:
            theCursor = self.cursorForPosition(theEvent.pos())
            self._followTag(theCursor)

        QTextEdit.mouseReleaseEvent(self, theEvent)
        self.docFooter.updateLineCount()

        return

    def resizeEvent(self, theEvent):
        """If the text editor is resize, we must make sure the document
        has its margins adjusted according to user preferences.
        """
        QTextEdit.resizeEvent(self, theEvent)
        self.updateDocMargins()
        return

    ##
    #  Signals and Slots
    ##

    @pyqtSlot(int, int, int)
    def _docChange(self, thePos, charsRemoved, charsAdded):
        """Triggered by QTextDocument->contentsChanged. This also
        triggers the syntax highlighter.
        """
        self.lastEdit = time()
        self.lastFind = None
        if self.qDocument.characterCount() > nwConst.MAX_DOCSIZE:
            self.theParent.makeAlert((
                "The document has grown too big and you cannot add more text to it. "
                "The maximum size of a single novelWriter document is %.2f\u202fMB."
            ) % (nwConst.MAX_DOCSIZE/1.0e6), nwAlert.ERROR)
            self.undo()
            return
        if not self.docChanged:
            self.setDocumentChanged(True)
        if not self.wcTimer.isActive():
            self.wcTimer.start()
        if self.doReplace and charsAdded == 1:
            self._docAutoReplace(self.qDocument.findBlock(thePos))
        return

    @pyqtSlot("QPoint")
    def _openContextMenu(self, thePos):
        """Triggered by right click to open the context menu. Also
        triggered by the Ctrl+. shortcut.
        """
        userCursor = self.textCursor()
        userSelection = userCursor.hasSelection()
        posCursor = self.cursorForPosition(thePos)

        mnuContext = QMenu()

        # Follow, Cut, Copy and Paste
        # ===========================

        if self._followTag(theCursor=posCursor, loadTag=False):
            mnuTag = QAction("Follow Tag", mnuContext)
            mnuTag.triggered.connect(lambda: self._followTag(theCursor=posCursor))
            mnuContext.addAction(mnuTag)
            mnuContext.addSeparator()

        if userSelection:
            mnuCut = QAction("Cut", mnuContext)
            mnuCut.triggered.connect(lambda: self.docAction(nwDocAction.CUT))
            mnuContext.addAction(mnuCut)

            mnuCopy = QAction("Copy", mnuContext)
            mnuCopy.triggered.connect(lambda: self.docAction(nwDocAction.COPY))
            mnuContext.addAction(mnuCopy)

        mnuPaste = QAction("Paste", mnuContext)
        mnuPaste.triggered.connect(lambda: self.docAction(nwDocAction.PASTE))
        mnuContext.addAction(mnuPaste)

        mnuContext.addSeparator()

        # Selections
        # ==========

        mnuSelAll = QAction("Select All", mnuContext)
        mnuSelAll.triggered.connect(lambda: self.docAction(nwDocAction.SEL_ALL))
        mnuContext.addAction(mnuSelAll)

        mnuSelWord = QAction("Select Word", mnuContext)
        mnuSelWord.triggered.connect(
            lambda: self._makePosSelection(QTextCursor.WordUnderCursor, thePos)
        )
        mnuContext.addAction(mnuSelWord)

        mnuSelPara = QAction("Select Paragraph", mnuContext)
        mnuSelPara.triggered.connect(
            lambda: self._makePosSelection(QTextCursor.BlockUnderCursor, thePos)
        )
        mnuContext.addAction(mnuSelPara)

        # Spell Checking
        # ==============

        posCursor  = self.cursorForPosition(thePos)
        spellCheck = self.spellCheck

        if posCursor.block().text().startswith("@"):
            spellCheck = False

        if spellCheck:
            posCursor.select(QTextCursor.WordUnderCursor)
            theWord = posCursor.selectedText().strip().strip(self.nonWord)
            spellCheck &= theWord != ""

        if spellCheck:
            logger.verbose("Looking up '%s' in the dictionary" % theWord)
            spellCheck &= not self.theDict.checkWord(theWord)

        if spellCheck:
            mnuContext.addSeparator()
            mnuHead = QAction("Spelling Suggestion(s)", mnuContext)
            mnuContext.addAction(mnuHead)

            theSuggest = self.theDict.suggestWords(theWord)[:15]
            if len(theSuggest) > 0:
                for aWord in theSuggest:
                    mnuWord = QAction("%s %s" % (nwUnicode.U_ENDASH, aWord), mnuContext)
                    mnuWord.triggered.connect(
                        lambda thePos, aWord=aWord : self._correctWord(posCursor, aWord)
                    )
                    mnuContext.addAction(mnuWord)
                mnuContext.addSeparator()
                mnuAdd = QAction("Add Word to Dictionary", mnuContext)
                mnuAdd.triggered.connect(lambda thePos : self._addWord(posCursor))
                mnuContext.addAction(mnuAdd)

            else:
                mnuHead = QAction("No Suggestions", mnuContext)
                mnuContext.addAction(mnuHead)

        # Open the context menu
        mnuContext.exec_(self.viewport().mapToGlobal(thePos))

        return

    @pyqtSlot("QTextCursor", str)
    def _correctWord(self, theCursor, theWord):
        """Slot for the spell check context menu triggering the
        replacement of a word with the word from the dictionary.
        """
        xPos = theCursor.selectionStart()
        theCursor.beginEditBlock()
        theCursor.removeSelectedText()
        theCursor.insertText(theWord)
        theCursor.endEditBlock()
        theCursor.setPosition(xPos)
        self.setTextCursor(theCursor)
        return

    @pyqtSlot("QTextCursor")
    def _addWord(self, theCursor):
        """Slot for the spell check context menu triggered when the user
        wants to add a word to the project dictionary.
        """
        theWord = theCursor.selectedText().strip().strip(self.nonWord)
        logger.debug("Added '%s' to project dictionary" % theWord)
        self.theDict.addWord(theWord)
        self.hLight.setDict(self.theDict)
        self.hLight.rehighlightBlock(theCursor.block())
        return

    @pyqtSlot()
    def _runCounter(self):
        """Decide whether to run the word counter, or not due to
        inactivity.
        """
        if self.theHandle is None:
            return

        if self.wCounter.isRunning():
            logger.verbose("Word counter is busy")
            return

        if time() - self.lastEdit < 5*self.wcInterval:
            logger.verbose("Running word counter")
            self.theParent.threadPool.start(self.wCounter)

        return

    @pyqtSlot(int, int, int)
    def _updateCounts(self, cCount, wCount, pCount):
        """Slot for the word counter's finished signal
        """
        theItem = self.nwDocument.getCurrentItem()
        if self.theHandle is None or theItem is None:
            return

        logger.verbose("Updating word count")

        self.charCount = cCount
        self.wordCount = wCount
        self.paraCount = pCount
        theItem.setCharCount(cCount)
        theItem.setWordCount(wCount)
        theItem.setParaCount(pCount)

        self.theParent.treeView.propagateCount(self.theHandle, wCount)
        self.theParent.treeView.projectWordCount()
        self.theParent.treeMeta.updateCounts(self.theHandle, cCount, wCount, pCount)
        self._checkDocSize(self.qDocument.characterCount())
        self.docFooter.updateCounts()

        return

    @pyqtSlot("QSizeF")
    def _docSizeChanged(self, theSize):
        """Called whenever the underlying document layout size changes.
        This is used to queue the repositioning of the cursor for very
        large documents to ensure the region where the cursor is being
        moved to has been drawn before the move is made.
        """
        if self.queuePos is not None:
            thePos = self.qDocument.documentLayout().hitTest(
                QPointF(theSize.width(), theSize.height()), Qt.FuzzyHit
            )
            if self.queuePos <= thePos:
                logger.verbose(
                    "Allowed cursor move to %d <= %d" % (self.queuePos, thePos)
                )
                self.setCursorPosition(self.queuePos)
                self.queuePos = None
            else:
                logger.verbose(
                    "Denied cursor move to %d > %d" % (self.queuePos, thePos)
                )
        return

    ##
    #  Internal Functions
    ##

    def _followTag(self, theCursor=None, loadTag=True):
        """Activated by Ctrl+Enter. Checks that we're in a block
        starting with '@'. We then find the word under the cursor and
        check that it is after the ':'. If all this is fine, we have a
        tag and can tell the document viewer to try and find and load
        the file where the tag is defined.
        """
        if theCursor is None:
            theCursor = self.textCursor()

        theBlock = theCursor.block()
        theText  = theBlock.text()

        if len(theText) == 0:
            return False

        if theText.startswith("@"):

            theCursor.select(QTextCursor.WordUnderCursor)
            theWord = theCursor.selectedText()
            cPos = theText.find(":")
            wPos = theCursor.selectionStart() - theBlock.position()
            if wPos <= cPos:
                return False

            if loadTag:
                logger.verbose("Attempting to follow tag '%s'" % theWord)
                self.theParent.docViewer.loadFromTag(theWord)
            else:
                logger.verbose("Potential tag '%s'" % theWord)

            return True

        return False

    def _openSpellContext(self):
        """Opens the spell check context menu at the current point of
        the cursor.
        """
        self._openContextMenu(self.cursorRect().center())
        return

    def _docAutoReplace(self, theBlock):
        """Autoreplace text elements based on main configuration.
        """
        if not theBlock.isValid():
            return

        theText   = theBlock.text()
        theCursor = self.textCursor()
        thePos    = theCursor.positionInBlock()
        theLen    = len(theText)

        if theLen < 1 or thePos-1 > theLen:
            return

        theOne   = theText[thePos-1:thePos]
        theTwo   = theText[thePos-2:thePos]
        theThree = theText[thePos-3:thePos]

        if self.mainConf.doReplaceDQuote and theTwo == " \"":
            theCursor.movePosition(QTextCursor.Left, QTextCursor.KeepAnchor, 1)
            theCursor.insertText(self.typDQOpen)

        elif self.mainConf.doReplaceDQuote and theOne == "\"":
            theCursor.movePosition(QTextCursor.Left, QTextCursor.KeepAnchor, 1)
            if thePos == 1:
                theCursor.insertText(self.typDQOpen)
            else:
                theCursor.insertText(self.typDQClose)

        elif self.mainConf.doReplaceSQuote and theTwo == " '":
            theCursor.movePosition(QTextCursor.Left, QTextCursor.KeepAnchor, 1)
            theCursor.insertText(self.typSQOpen)

        elif self.mainConf.doReplaceSQuote and theOne == "'":
            theCursor.movePosition(QTextCursor.Left, QTextCursor.KeepAnchor, 1)
            if thePos == 1:
                theCursor.insertText(self.typSQOpen)
            else:
                theCursor.insertText(self.typSQClose)

        elif self.mainConf.doReplaceDash and theThree == "---":
            theCursor.movePosition(QTextCursor.Left, QTextCursor.KeepAnchor, 3)
            theCursor.insertText(nwUnicode.U_EMDASH)

        elif self.mainConf.doReplaceDash and theTwo == "--":
            theCursor.movePosition(QTextCursor.Left, QTextCursor.KeepAnchor, 2)
            theCursor.insertText(nwUnicode.U_ENDASH)

        elif self.mainConf.doReplaceDash and theTwo == nwUnicode.U_ENDASH+"-":
            theCursor.movePosition(QTextCursor.Left, QTextCursor.KeepAnchor, 2)
            theCursor.insertText(nwUnicode.U_EMDASH)

        elif self.mainConf.doReplaceDots and theThree == "...":
            theCursor.movePosition(QTextCursor.Left, QTextCursor.KeepAnchor, 3)
            theCursor.insertText(nwUnicode.U_HELLIP)

        return

    def _replaceQuotes(self, sQuote, oQuote, cQuote):
        """Replace all straight quotes in the selected text.
        """
        theCursor = self.textCursor()
        if theCursor.hasSelection():
            posS = theCursor.selectionStart()
            posE = theCursor.selectionEnd()
            closeCheck = (
                " ", "\n", nwUnicode.U_LSEP, nwUnicode.U_PSEP
            )

            self._allowAutoReplace(False)
            for posC in range(posS, posE+1):
                theCursor.setPosition(posC)
                theCursor.movePosition(QTextCursor.Left, QTextCursor.KeepAnchor, 2)
                selText = theCursor.selectedText()

                nS = len(selText)
                if nS == 2:
                    pC = selText[0]
                    cC = selText[1]
                elif nS == 1:
                    pC = " "
                    cC = selText[0]
                else:
                    continue

                if cC != sQuote:
                    continue

                theCursor.clearSelection()
                theCursor.setPosition(posC)
                if pC in closeCheck:
                    theCursor.beginEditBlock()
                    theCursor.movePosition(QTextCursor.Left, QTextCursor.KeepAnchor, 1)
                    theCursor.insertText(oQuote)
                    theCursor.endEditBlock()
                else:
                    theCursor.beginEditBlock()
                    theCursor.movePosition(QTextCursor.Left, QTextCursor.KeepAnchor, 1)
                    theCursor.insertText(cQuote)
                    theCursor.endEditBlock()

            self._allowAutoReplace(True)

        else:
            self.theParent.makeAlert(
                "Please selection some text before calling replace quotes.", nwAlert.ERROR
            )

        return

    def _checkDocSize(self, theSize):
        """Check if document size crosses the big document limit set in
        config. If so, we will set the big document flag to True.
        """
        bigLim = self.mainConf.bigDocLimit*1000
        newState = theSize > bigLim

        if newState != self.bigDoc:
            if newState:
                logger.info(
                    f"The document size is {theSize:n} > {bigLim:n}, "
                    f"big doc mode has been enabled"
                )
            else:
                logger.info(
                    f"The document size is {theSize:n} <= {bigLim:n}, "
                    f"big doc mode has been disabled"
                )

        self.bigDoc = newState

        return

    def _wrapSelection(self, tBefore, tAfter=None):
        """Wraps the selected text in whatever is in tBefore and tAfter.
        If there is no selection, the autoSelect setting decides the
        action. AutoSelect will select the word under the cursor before
        wrapping it. If this feature is disabled, nothing is done.
        """
        if tAfter is None:
            tAfter = tBefore

        theCursor = self._autoSelect()
        if theCursor.hasSelection():
            posS = theCursor.selectionStart()
            posE = theCursor.selectionEnd()

            blockS = self.qDocument.findBlock(posS)
            blockE = self.qDocument.findBlock(posE)
            if blockS != blockE:
                posE = blockS.position() + blockS.length() - 1

            theCursor.clearSelection()
            theCursor.beginEditBlock()
            theCursor.setPosition(posE)
            theCursor.insertText(tAfter)
            theCursor.setPosition(posS)
            theCursor.insertText(tBefore)
            theCursor.endEditBlock()

            theCursor.setPosition(posE + len(tBefore), QTextCursor.MoveAnchor)
            theCursor.setPosition(posS + len(tBefore), QTextCursor.KeepAnchor)
            self.setTextCursor(theCursor)

        else:
            logger.warning("No selection made, nothing to do")
        return

    def _clearSurrounding(self, theCursor, nChars):
        """Clears n characters before and after the cursor.
        """
        if theCursor.hasSelection():
            posS = theCursor.selectionStart()
            posE = theCursor.selectionEnd()
            theCursor.clearSelection()
            theCursor.beginEditBlock()
            theCursor.setPosition(posS)
            for i in range(nChars):
                theCursor.deletePreviousChar()
            theCursor.setPosition(posE)
            for i in range(nChars):
                theCursor.deletePreviousChar()
            theCursor.endEditBlock()
            theCursor.clearSelection()
        else:
            logger.warning("No selection made, nothing to do")
        return

    def _autoSelect(self):
        """Returns a cursor which may or may not have a selection based
        on user settings and document action.
        """
        theCursor = self.textCursor()
        if self.mainConf.autoSelect and not theCursor.hasSelection():
            theCursor.select(QTextCursor.WordUnderCursor)
            posS = theCursor.selectionStart()
            posE = theCursor.selectionEnd()

            # Underscore counts as a part of the word, so check that the
            # selection isn't wrapped in italics markers.
            reSelect = False
            if self.qDocument.characterAt(posS) == "_":
                posS += 1
                reSelect = True
            if self.qDocument.characterAt(posE) == "_":
                posE -= 1
                reSelect = True
            if reSelect:
                theCursor.clearSelection()
                theCursor.setPosition(posS, QTextCursor.MoveAnchor)
                theCursor.setPosition(posE-1, QTextCursor.KeepAnchor)

            self.setTextCursor(theCursor)

        return theCursor

    def _toggleFormat(self, fLen, fChar):
        """Toggle the formatting of a specific type for a piece of text.
        If more than one block is selected, the formatting is applied to
        the first block.
        """
        theCursor = self._autoSelect()
        if theCursor.hasSelection():
            posS = theCursor.selectionStart()
            posE = theCursor.selectionEnd()

            blockS = self.qDocument.findBlock(posS)
            blockE = self.qDocument.findBlock(posE)

            if blockS != blockE:
                posE = blockS.position() + blockS.length() - 1
                theCursor.clearSelection()
                theCursor.setPosition(posS, QTextCursor.MoveAnchor)
                theCursor.setPosition(posE, QTextCursor.KeepAnchor)
                self.setTextCursor(theCursor)

            numB = 0
            for n in range(fLen):
                if self.qDocument.characterAt(posS-n-1) == fChar:
                    numB += 1
                else:
                    break

            numA = 0
            for n in range(fLen):
                if self.qDocument.characterAt(posE+n) == fChar:
                    numA += 1
                else:
                    break

            cLevel = min(numB, numA)
            if cLevel == fLen:
                self._clearSurrounding(theCursor, fLen)
            else:
                self._wrapSelection(fChar*fLen)

        return

    def _formatBlock(self, docAction):
        """Changes the block format of the block under the cursor.
        """
        theCursor = self.textCursor()
        theBlock = theCursor.block()
        if not theBlock.isValid():
            logger.debug("Invalid block selected for action %s" % str(docAction))
            return False

        theText = theBlock.text()
        if len(theText.strip()) == 0:
            logger.debug("Empty block selected for action %s" % str(docAction))
            return False

        # Remove existing format first, if any
        if theText.startswith("@"):
            logger.error("Cannot apply block format to keyword/value line")
            return False
        elif theText.startswith("% "):
            newText = theText[2:]
            cOffset = 2
        elif theText.startswith("%"):
            newText = theText[1:]
            cOffset = 1
        elif theText.startswith("# "):
            newText = theText[2:]
            cOffset = 2
        elif theText.startswith("## "):
            newText = theText[3:]
            cOffset = 3
        elif theText.startswith("### "):
            newText = theText[4:]
            cOffset = 4
        elif theText.startswith("#### "):
            newText = theText[5:]
            cOffset = 5
        else:
            newText = theText
            cOffset = 0

        # Apply new format
        if docAction == nwDocAction.BLOCK_COM:
            theText = "% "+newText
            cOffset -= 2
        elif docAction == nwDocAction.BLOCK_H1:
            theText = "# "+newText
            cOffset -= 2
        elif docAction == nwDocAction.BLOCK_H2:
            theText = "## "+newText
            cOffset -= 3
        elif docAction == nwDocAction.BLOCK_H3:
            theText = "### "+newText
            cOffset -= 4
        elif docAction == nwDocAction.BLOCK_H4:
            theText = "#### "+newText
            cOffset -= 5
        elif docAction == nwDocAction.BLOCK_TXT:
            theText = newText
            cOffset -= 0
        else:
            logger.error(
                "Unknown or unsupported block format requested: %s" % str(docAction)
            )
            return False

        # Replace the block text
        theCursor.beginEditBlock()
        posO = theCursor.position()
        theCursor.select(QTextCursor.BlockUnderCursor)
        posS = theCursor.selectionStart()
        theCursor.removeSelectedText()
        theCursor.setPosition(posS)
        if posS > 0:
            theCursor.insertBlock()
        theCursor.insertText(theText)
        if posO - cOffset >= 0:
            theCursor.setPosition(posO - cOffset)
        theCursor.endEditBlock()
        self.setTextCursor(theCursor)

        return True

    def _makeSelection(self, selMode):
        """Wrapper function to select text based on a selection mode.
        """
        theCursor = self.textCursor()
        theCursor.clearSelection()
        theCursor.select(selMode)

        if selMode == QTextCursor.WordUnderCursor:
            theCursor = self._autoSelect()

        elif selMode == QTextCursor.BlockUnderCursor:
            # This selection mode also selects the preceding oaragraph
            # separator, which we want to avoid.
            posS = theCursor.selectionStart()
            posE = theCursor.selectionEnd()
            selTxt = theCursor.selectedText()
            if selTxt.startswith(nwUnicode.U_PSEP):
                theCursor.setPosition(posS+1, QTextCursor.MoveAnchor)
                theCursor.setPosition(posE, QTextCursor.KeepAnchor)

        self.setTextCursor(theCursor)

        return

    def _makePosSelection(self, selMode, thePos):
        """Wrapper function to select text based on selection mode, but
        first move cursor to given position.
        """
        theCursor = self.cursorForPosition(thePos)
        self.setTextCursor(theCursor)
        self._makeSelection(selMode)
        return

    def _beginSearch(self):
        """Sets the selected text as the search text for the search bar.
        """
        theCursor = self.textCursor()
        if theCursor.hasSelection():
            self.docSearch.setSearchText(theCursor.selectedText())
        else:
            self.docSearch.setSearchText(None)
        self.updateDocMargins()
        return

    def _beginReplace(self):
        """Opens the replace line of the search bar and sets the find
        text if a selection has been made, and resets the replace text.
        """
        theCursor = self.textCursor()
        if theCursor.hasSelection():
            self.docSearch.setSearchText(theCursor.selectedText())
        else:
            self.docSearch.setSearchText(None)
        self.docSearch.setReplaceText("")
        self.updateDocMargins()
        return

    def _findNext(self, isBackward=False):
        """Searches for the next or previous occurrence of the search
        bar text in the document. Wraps around if not found and loop is
        enabled, or continues to next file if next file is enabled.
        """
        if not self.docSearch.isVisible():
            self._beginSearch()
            return

        findOpt = QTextDocument.FindFlag(0)
        if isBackward:
            findOpt |= QTextDocument.FindBackward
        if self.docSearch.isCaseSense:
            findOpt |= QTextDocument.FindCaseSensitively
        if self.docSearch.isWholeWord:
            findOpt |= QTextDocument.FindWholeWords

        searchFor = self.docSearch.getSearchObject()
        wasFound  = self.find(searchFor, findOpt)
        if not wasFound:
            if self.docSearch.doNextFile and not isBackward:
                self.theParent.openNextDocument(
                    self.theHandle, wrapAround=self.docSearch.doLoop
                )
            elif self.docSearch.doLoop:
                theCursor = self.textCursor()
                theCursor.movePosition(
                    QTextCursor.End if isBackward else QTextCursor.Start
                )
                self.setTextCursor(theCursor)
                wasFound = self.find(searchFor, findOpt)

        if wasFound:
            theCursor = self.textCursor()
            self.lastFind = (theCursor.selectionStart(), theCursor.selectionEnd())

        return

    def _replaceNext(self):
        """Searches for the next occurrence of the search bar text in
        the document and replaces it with the replace text. Calls search
        next automatically when done.
        """
        if not self.docSearch.isVisible():
            # The search tool is not active, so we activate it.
            self._beginSearch()
            return

        theCursor = self.textCursor()
        if not theCursor.hasSelection():
            # We have no text selected at all, so just make this a
            # regular find next call.
            self._findNext()
            return

        if self.lastFind is None and theCursor.hasSelection():
            # If we have a selection but no search, it may have been the
            # text we triggered the search with, in which case we search
            # again from the beginning of that selection to make sure we
            # have a valid result.
            sPos = theCursor.selectionStart()
            theCursor.clearSelection()
            theCursor.setPosition(sPos)
            self.setTextCursor(theCursor)
            self._findNext()
            theCursor = self.textCursor()

        if self.lastFind is None:
            # In case the above didn't find a result, we give up here.
            return

        searchFor = self.docSearch.getSearchText()
        replWith  = self.docSearch.getReplaceText()

        if self.docSearch.doMatchCap:
            replWith = transferCase(theCursor.selectedText(), replWith)

        # Make sure the selected text was selected by an actual find
        # call, and not the user.
        try:
            isFind  = self.lastFind[0] == theCursor.selectionStart()
            isFind &= self.lastFind[1] == theCursor.selectionEnd()
        except Exception:
            isFind = False

        if isFind:
            theCursor.beginEditBlock()
            theCursor.removeSelectedText()
            theCursor.insertText(replWith)
            theCursor.endEditBlock()
            theCursor.setPosition(theCursor.selectionEnd())
            self.setTextCursor(theCursor)
            logger.verbose("Replaced occurrence of '%s' with '%s' on line %d" % (
                searchFor, replWith, theCursor.blockNumber()
            ))
        else:
            logger.error("The selected text is not a search result, skipping replace")

        self._findNext()

        return

    def _setupSpellChecking(self):
        """Create the spell checking object based on the spellTool
        setting in config.
        """
        if self.mainConf.spellTool == nwConst.SP_ENCHANT:
            from nw.core.spellcheck import NWSpellEnchant
            self.theDict = NWSpellEnchant()
        else:
            self.theDict = NWSpellSimple()

        self.hLight.setDict(self.theDict)

        return

    def _allowAutoReplace(self, theState):
        """used to enable/disable the auto-replace feature temporarily.
        """
        if theState:
            self.doReplace = self.mainConf.doReplace
        else:
            self.doReplace = False
        return

# END Class GuiDocEditor

# =============================================================================================== #
#  The Off-GUI Thread Word Counter
#  A runnable for the word counter to be run in the thread pool off the main GUI thread.
# =============================================================================================== #

class BackgroundWordCounter(QRunnable):

    def __init__(self, docEditor):
        QRunnable.__init__(self)
        self.docEditor = docEditor
        self.signals = BackgroundWordCounterSignals()
        self._isRunning = False
        return

    def isRunning(self):
        return self._isRunning

    @pyqtSlot()
    def run(self):
        """Overloaded run function for the word counter, forwarding the
        call to the function that does the actual counting.
        """
        self._isRunning = True
        theText = self.docEditor.getText()
        cC, wC, pC = countWords(theText)
        self.signals.countsReady.emit(cC, wC, pC)
        self._isRunning = False
        return

## END Class BackgroundWordCounter

class BackgroundWordCounterSignals(QObject):

    countsReady = pyqtSignal(int, int, int)

# END Class BackgroundWordCounterSignals

# =============================================================================================== #
#  The Embedded Document Search/Replace Feature
#  Only used by DocEditor, and is at a fixed position in the QTextEdit's viewport
# =============================================================================================== #

class GuiDocEditSearch(QFrame):

    def __init__(self, docEditor):
        QFrame.__init__(self, docEditor)

        logger.debug("Initialising GuiDocEditSearch ...")

        self.mainConf   = nw.CONFIG
        self.docEditor  = docEditor
        self.theParent  = docEditor.theParent
        self.theProject = docEditor.theProject
        self.theTheme   = docEditor.theTheme

        self.repVisible  = False
        self.isCaseSense = self.mainConf.searchCase
        self.isWholeWord = self.mainConf.searchWord
        self.isRegEx     = self.mainConf.searchRegEx
        self.doLoop      = self.mainConf.searchLoop
        self.doNextFile  = self.mainConf.searchNextFile
        self.doMatchCap  = self.mainConf.searchMatchCap

        mPx = self.mainConf.pxInt(6)
        tPx = int(0.8*self.theTheme.fontPixelSize)
        boxFont = self.theTheme.guiFont
        boxFont.setPointSizeF(0.9*self.theTheme.fontPointSize)

        self.setContentsMargins(0, 0, 0, 0)
        self.setAutoFillBackground(True)
        self.setFrameStyle(QFrame.StyledPanel | QFrame.Plain)

        self.mainBox = QGridLayout(self)
        self.setLayout(self.mainBox)

        # Text Boxes
        # ==========
        self.searchBox = QLineEdit(self)
        self.searchBox.setFont(boxFont)
        self.searchBox.setPlaceholderText("Search")
        self.searchBox.returnPressed.connect(self._doSearch)

        self.replaceBox = QLineEdit(self)
        self.replaceBox.setFont(boxFont)
        self.replaceBox.setPlaceholderText("Replace")
        self.replaceBox.returnPressed.connect(self._doReplace)

        self.searchOpt = QToolBar(self)
        self.searchOpt.setToolButtonStyle(Qt.ToolButtonIconOnly)
        self.searchOpt.setIconSize(QSize(tPx, tPx))
        self.searchOpt.setContentsMargins(0, 0, 0, 0)
        self.searchOpt.setStyleSheet(r"QToolBar {padding: 0;}")

        self.searchLabel = QLabel("Search")
        self.searchLabel.setFont(boxFont)
        self.searchLabel.setIndent(self.mainConf.pxInt(6))

        self.toggleCase = QAction("Case Sensitive", self)
        self.toggleCase.setToolTip("Match case")
        self.toggleCase.setIcon(self.theTheme.getIcon("search_case"))
        self.toggleCase.setCheckable(True)
        self.toggleCase.setChecked(self.isCaseSense)
        self.toggleCase.toggled.connect(self._doToggleCase)
        self.searchOpt.addAction(self.toggleCase)

        self.toggleWord = QAction("Whole Words Only", self)
        self.toggleWord.setToolTip("Match whole words")
        self.toggleWord.setIcon(self.theTheme.getIcon("search_word"))
        self.toggleWord.setCheckable(True)
        self.toggleWord.setChecked(self.isWholeWord)
        self.toggleWord.toggled.connect(self._doToggleWord)
        self.searchOpt.addAction(self.toggleWord)

        self.toggleRegEx = QAction("RegEx Mode", self)
        self.toggleRegEx.setToolTip("Use regular expressions (requires Qt 5.3)")
        self.toggleRegEx.setIcon(self.theTheme.getIcon("search_regex"))
        self.toggleRegEx.setCheckable(True)
        self.toggleRegEx.setChecked(self.isRegEx)
        self.toggleRegEx.toggled.connect(self._doToggleRegEx)
        self.searchOpt.addAction(self.toggleRegEx)

        self.toggleLoop = QAction("Loop Search", self)
        self.toggleLoop.setToolTip("Loop the search when reaching the end")
        self.toggleLoop.setIcon(self.theTheme.getIcon("search_loop"))
        self.toggleLoop.setCheckable(True)
        self.toggleLoop.setChecked(self.doLoop)
        self.toggleLoop.toggled.connect(self._doToggleLoop)
        self.searchOpt.addAction(self.toggleLoop)

        self.toggleProject = QAction("Search Next File", self)
        self.toggleProject.setToolTip("Continue searching in the next file")
        self.toggleProject.setIcon(self.theTheme.getIcon("search_project"))
        self.toggleProject.setCheckable(True)
        self.toggleProject.setChecked(self.doNextFile)
        self.toggleProject.toggled.connect(self._doToggleProject)
        self.searchOpt.addAction(self.toggleProject)

        self.searchOpt.addSeparator()

        self.toggleMatchCap = QAction("Preserve Case", self)
        self.toggleMatchCap.setToolTip("Preserve case on replace")
        self.toggleMatchCap.setIcon(self.theTheme.getIcon("search_preserve"))
        self.toggleMatchCap.setCheckable(True)
        self.toggleMatchCap.setChecked(self.doMatchCap)
        self.toggleMatchCap.toggled.connect(self._doToggleMatchCap)
        self.searchOpt.addAction(self.toggleMatchCap)

        self.searchOpt.addSeparator()

        self.cancelSearch = QAction("Close Search", self)
        self.cancelSearch.setToolTip("Close the search box [Esc]")
        self.cancelSearch.setIcon(self.theTheme.getIcon("search_cancel"))
        self.cancelSearch.triggered.connect(self._doClose)
        self.searchOpt.addAction(self.cancelSearch)

        # Buttons
        # =======
        bPx = self.searchBox.sizeHint().height()

        self.showReplace = QToolButton(self)
        self.showReplace.setArrowType(Qt.RightArrow)
        self.showReplace.setCheckable(True)
        self.showReplace.setToolTip("Show/hide the replace text box")
        self.showReplace.setStyleSheet(r"QToolButton {border: none; background: transparent;}")
        self.showReplace.toggled.connect(self._doToggleReplace)

        self.searchButton = QPushButton(self.theTheme.getIcon("search"), "")
        self.searchButton.setFixedSize(QSize(bPx, bPx))
        self.searchButton.setToolTip("Find in current document")
        self.searchButton.clicked.connect(self._doSearch)

        self.replaceButton = QPushButton(self.theTheme.getIcon("search-replace"), "")
        self.replaceButton.setFixedSize(QSize(bPx, bPx))
        self.replaceButton.setToolTip("Find and replace in current document")
        self.replaceButton.clicked.connect(self._doReplace)

        self.mainBox.addWidget(self.searchLabel,   0, 0, 1, 2, Qt.AlignLeft)
        self.mainBox.addWidget(self.searchOpt,     0, 2, 1, 2, Qt.AlignRight)
        self.mainBox.addWidget(self.showReplace,   1, 0, 1, 1)
        self.mainBox.addWidget(self.searchBox,     1, 1, 1, 2)
        self.mainBox.addWidget(self.searchButton,  1, 3, 1, 1)
        self.mainBox.addWidget(self.replaceBox,    2, 1, 1, 2)
        self.mainBox.addWidget(self.replaceButton, 2, 3, 1, 1)

        self.mainBox.setColumnStretch(0, 1)
        self.mainBox.setColumnStretch(1, 0)
        self.mainBox.setColumnStretch(2, 0)
        self.mainBox.setColumnStretch(3, 0)
        self.mainBox.setColumnStretch(4, 0)
        self.mainBox.setSpacing(self.mainConf.pxInt(2))
        self.mainBox.setContentsMargins(mPx, mPx, mPx, mPx)

        boxWidth = self.mainConf.pxInt(200)
        self.searchBox.setFixedWidth(boxWidth)
        self.replaceBox.setFixedWidth(boxWidth)
        self.replaceBox.setVisible(False)
        self.replaceButton.setVisible(False)
        self.adjustSize()

        # Construct Box Colours
        qPalette = self.searchBox.palette()
        baseCol = qPalette.base().color()
        rCol = baseCol.redF()   + 0.1
        gCol = baseCol.greenF() - 0.1
        bCol = baseCol.blueF()  - 0.1

        mCol = max(rCol, gCol, bCol, 1.0)
        errCol = QColor()
        errCol.setRedF(rCol/mCol)
        errCol.setGreenF(gCol/mCol)
        errCol.setBlueF(bCol/mCol)

        self.rxCol = {
            True  : baseCol,
            False : errCol
        }

        logger.debug("GuiDocEditSearch initialisation complete")

        return

    def closeSearch(self):
        """Close the search box.
        """
        self.mainConf.searchCase     = self.isCaseSense
        self.mainConf.searchWord     = self.isWholeWord
        self.mainConf.searchRegEx    = self.isRegEx
        self.mainConf.searchLoop     = self.doLoop
        self.mainConf.searchNextFile = self.doNextFile
        self.mainConf.searchMatchCap = self.doMatchCap

        self.showReplace.setChecked(False)
        self.setVisible(False)
        self.docEditor.updateDocMargins()
        self.docEditor.setFocus()

        return

    def cycleFocus(self, toNext):
        """The tab key just alternates focus between the two input
        boxes, if the replace box is visible.
        """
        if self.replaceBox.isVisible():
            if self.searchBox.hasFocus():
                self.replaceBox.setFocus(True)
                return True
            elif self.replaceBox.hasFocus():
                self.searchBox.setFocus(True)
                return True
        return False

    def anyFocus(self):
        """Returns true if any of the input boxes have focus.
        """
        return self.searchBox.hasFocus() | self.replaceBox.hasFocus()

    ##
    #  Get and Set Functions
    ##

    def setSearchText(self, theText):
        """Open the search bar and set the search text to the text
        provided, if any.
        """
        if not self.isVisible():
            self.setVisible(True)
        if theText is not None:
            self.searchBox.setText(theText)
        self.searchBox.setFocus()
        if self.isRegEx:
            self._alertSearchValid(True)
        logger.verbose("Setting search text to '%s'" % theText)
        return True

    def setReplaceText(self, theText):
        """Set the replace text.
        """
        self.showReplace.setChecked(True)
        self.replaceBox.setFocus()
        self.replaceBox.setText(theText)
        return True

    def getSearchObject(self):
        """Return the current search text either as text or as a regular
        expression object.
        """
        theText = self.searchBox.text()
        if self.isRegEx:
            # Using the Unicode-capable QRegularExpression class was
            # only added in Qt 5.13. Otherwise, 5.3 and up supports
            # only the QRegExp class.
            if self.mainConf.verQtValue >= 51300:
                rxOpt = QRegularExpression.UseUnicodePropertiesOption
                if not self.isCaseSense:
                    rxOpt |= QRegularExpression.CaseInsensitiveOption
                theRegEx = QRegularExpression(theText, rxOpt)
                self._alertSearchValid(theRegEx.isValid())
                return theRegEx

            elif self.mainConf.verQtValue >= 50300:
                if self.isCaseSense:
                    rxOpt = Qt.CaseSensitive
                else:
                    rxOpt = Qt.CaseInsensitive
                theRegEx = QRegExp(theText, rxOpt)
                self._alertSearchValid(theRegEx.isValid())
                return theRegEx

        return theText

    def getSearchText(self):
        """Return the current search text.
        """
        return self.searchBox.text()

    def getReplaceText(self):
        """Return the current replace text.
        """
        return self.replaceBox.text()

    ##
    #  Slots
    ##

    def _doClose(self):
        """Hide the search/replace bar.
        """
        self.closeSearch()
        return

    def _doSearch(self):
        """Call the search action function for the document editor.
        """
        modKey = qApp.keyboardModifiers()
        if modKey == Qt.ShiftModifier:
            self.docEditor.docAction(nwDocAction.GO_PREV)
        else:
            self.docEditor.docAction(nwDocAction.GO_NEXT)
        return

    def _doReplace(self):
        """Call the replace action function for the document editor.
        """
        self.docEditor.docAction(nwDocAction.REPL_NEXT)
        return

    def _doToggleReplace(self, theState):
        """Toggle the show/hide of the
        """
        if theState:
            self.showReplace.setArrowType(Qt.DownArrow)
        else:
            self.showReplace.setArrowType(Qt.RightArrow)
        self.replaceBox.setVisible(theState)
        self.replaceButton.setVisible(theState)
        self.repVisible = theState
        self.adjustSize()
        self.docEditor.updateDocMargins()
        return

    def _doToggleCase(self, theState):
        """Enable/disable case sensitive mode.
        """
        self.isCaseSense = theState
        return

    def _doToggleWord(self, theState):
        """Enable/disable whole word search mode.
        """
        self.isWholeWord = theState
        return

    def _doToggleRegEx(self, theState):
        """Enable/disable regular expression search mode.
        """
        self.isRegEx = theState
        return

    def _doToggleLoop(self, theState):
        """Enable/disable looping the search.
        """
        self.doLoop = theState
        return

    def _doToggleProject(self, theState):
        """Enable/disable continuing search in next project file.
        """
        self.doNextFile = theState
        return

    def _doToggleMatchCap(self, theState):
        """Enable/disable preserving capitalisation when replacing.
        """
        self.doMatchCap = theState
        return

    ##
    #  Internal Functions
    ##

    def _alertSearchValid(self, isValid):
        """Highlight the search box to indicate the search string is or
        isn't valid. Take the colour from the replace box.
        """
        qPalette = self.replaceBox.palette()
        qPalette.setColor(QPalette.Base, self.rxCol[isValid])
        self.searchBox.setPalette(qPalette)
        return

# END Class GuiDocEditSearch

# =============================================================================================== #
#  The Embedded Document Header
#  Only used by DocEditor, and is at a fixed position in the QTextEdit's viewport
# =============================================================================================== #

class GuiDocEditHeader(QWidget):

    def __init__(self, docEditor):
        QWidget.__init__(self, docEditor)

        logger.debug("Initialising GuiDocEditHeader ...")

        self.mainConf   = nw.CONFIG
        self.docEditor  = docEditor
        self.theParent  = docEditor.theParent
        self.theProject = docEditor.theProject
        self.theTheme   = docEditor.theTheme
        self.theHandle  = None

        fPx = int(0.9*self.theTheme.fontPixelSize)
        hSp = self.mainConf.pxInt(6)
        self.buttonSize = fPx + hSp

        # Main Widget Settings
        self.setAutoFillBackground(True)

        # Title Label
        self.theTitle = QLabel()
        self.theTitle.setText("")
        self.theTitle.setIndent(0)
        self.theTitle.setMargin(0)
        self.theTitle.setContentsMargins(0, 0, 0, 0)
        self.theTitle.setAutoFillBackground(True)
        self.theTitle.setAlignment(Qt.AlignHCenter | Qt.AlignTop)
        self.theTitle.setFixedHeight(fPx)

        lblFont = self.theTitle.font()
        lblFont.setPointSizeF(0.9*self.theTheme.fontPointSize)
        self.theTitle.setFont(lblFont)

        buttonStyle = (
            "QToolButton {{border: none; background: transparent;}} "
            "QToolButton:hover {{border: none; background: rgba({0},{1},{2},0.2);}}"
        ).format(*self.theTheme.colText)

        # Buttons
        self.editButton = QToolButton(self)
        self.editButton.setIcon(self.theTheme.getIcon("edit"))
        self.editButton.setContentsMargins(0, 0, 0, 0)
        self.editButton.setIconSize(QSize(fPx, fPx))
        self.editButton.setFixedSize(fPx, fPx)
        self.editButton.setStyleSheet(buttonStyle)
        self.editButton.setToolButtonStyle(Qt.ToolButtonIconOnly)
        self.editButton.setVisible(False)
        self.editButton.setToolTip("Edit document meta")
        self.editButton.clicked.connect(self._editDocument)

        self.searchButton = QToolButton(self)
        self.searchButton.setIcon(self.theTheme.getIcon("search"))
        self.searchButton.setContentsMargins(0, 0, 0, 0)
        self.searchButton.setIconSize(QSize(fPx, fPx))
        self.searchButton.setFixedSize(fPx, fPx)
        self.searchButton.setStyleSheet(buttonStyle)
        self.searchButton.setToolButtonStyle(Qt.ToolButtonIconOnly)
        self.searchButton.setVisible(False)
        self.searchButton.setToolTip("Search document")
        self.searchButton.clicked.connect(self._searchDocument)

        self.minmaxButton = QToolButton(self)
        self.minmaxButton.setIcon(self.theTheme.getIcon("maximise"))
        self.minmaxButton.setContentsMargins(0, 0, 0, 0)
        self.minmaxButton.setIconSize(QSize(fPx, fPx))
        self.minmaxButton.setFixedSize(fPx, fPx)
        self.minmaxButton.setStyleSheet(buttonStyle)
        self.minmaxButton.setToolButtonStyle(Qt.ToolButtonIconOnly)
        self.minmaxButton.setVisible(False)
        self.minmaxButton.setToolTip("Toggle Focus Mode")
        self.minmaxButton.clicked.connect(self._minmaxDocument)

        self.closeButton = QToolButton(self)
        self.closeButton.setIcon(self.theTheme.getIcon("close"))
        self.closeButton.setContentsMargins(0, 0, 0, 0)
        self.closeButton.setIconSize(QSize(fPx, fPx))
        self.closeButton.setFixedSize(fPx, fPx)
        self.closeButton.setStyleSheet(buttonStyle)
        self.closeButton.setToolButtonStyle(Qt.ToolButtonIconOnly)
        self.closeButton.setVisible(False)
        self.closeButton.setToolTip("Close the document")
        self.closeButton.clicked.connect(self._closeDocument)

        # Assemble Layout
        self.outerBox = QHBoxLayout()
        self.outerBox.setSpacing(hSp)
        self.outerBox.addWidget(self.editButton, 0)
        self.outerBox.addWidget(self.searchButton, 0)
        self.outerBox.addWidget(self.theTitle, 1)
        self.outerBox.addWidget(self.minmaxButton, 0)
        self.outerBox.addWidget(self.closeButton, 0)
        self.setLayout(self.outerBox)

        # Fix the Colours
        self.matchColours()

        logger.debug("GuiDocEditHeader initialisation complete")

        return

    ##
    #  Methods
    ##

    def matchColours(self):
        """Update the colours of the widget to match those of the syntax
        theme rather than the main GUI.
        """
        thePalette = QPalette()
        thePalette.setColor(QPalette.Window, QColor(*self.theTheme.colBack))
        thePalette.setColor(QPalette.WindowText, QColor(*self.theTheme.colText))
        thePalette.setColor(QPalette.Text, QColor(*self.theTheme.colText))

        self.setPalette(thePalette)
        self.theTitle.setPalette(thePalette)

        return

    def setTitleFromHandle(self, tHandle):
        """Sets the document title from the handle, or alternatively,
        set the whole document path.
        """
        self.theHandle = tHandle
        if tHandle is None:
            self.theTitle.setText("")
            self.editButton.setVisible(False)
            self.searchButton.setVisible(False)
            self.closeButton.setVisible(False)
            self.minmaxButton.setVisible(False)
            return True

        if self.mainConf.showFullPath:
            tTitle = []
            tTree = self.theProject.projTree.getItemPath(tHandle)
            for aHandle in reversed(tTree):
                nwItem = self.theProject.projTree[aHandle]
                if nwItem is not None:
                    tTitle.append(nwItem.itemName)
            sSep = "  %s  " % nwUnicode.U_RSAQUO
            self.theTitle.setText(sSep.join(tTitle))
        else:
            nwItem = self.theProject.projTree[tHandle]
            if nwItem is None:
                return False
            self.theTitle.setText(nwItem.itemName)

        self.editButton.setVisible(True)
        self.searchButton.setVisible(True)
        self.closeButton.setVisible(True)
        self.minmaxButton.setVisible(True)

        return True

    ##
    #  Slots
    ##

    def _editDocument(self):
        """Open the edit item dialog from the main GUI.
        """
        self.theParent.editItem(self.theHandle)
        return

    def _searchDocument(self):
        """Toggle the visibility of the search box.
        """
        self.docEditor.toggleSearch()
        return

    def _closeDocument(self):
        """Trigger the close editor on the main window.
        """
        self.theParent.closeDocEditor()
        self.editButton.setVisible(False)
        self.searchButton.setVisible(False)
        self.closeButton.setVisible(False)
        self.minmaxButton.setVisible(False)
        return

    def _minmaxDocument(self):
        """Switch on or off Focus Mode.
        """
        self.theParent.toggleFocusMode()
        if self.theParent.isFocusMode:
            self.minmaxButton.setIcon(self.theTheme.getIcon("minimise"))
            self.setContentsMargins(self.buttonSize, 0, 0, 0)
            self.editButton.setVisible(False)
            self.searchButton.setVisible(False)
            self.closeButton.setVisible(False)
        else:
            self.minmaxButton.setIcon(self.theTheme.getIcon("maximise"))
            self.setContentsMargins(0, 0, 0, 0)
            self.editButton.setVisible(True)
            self.searchButton.setVisible(True)
            self.closeButton.setVisible(True)
        return

    ##
    #  Events
    ##

    def mousePressEvent(self, theEvent):
        """Capture a click on the title and ensure that the item is
        selected in the project tree.
        """
        self.theParent.treeView.setSelectedHandle(self.theHandle, doScroll=True)
        return

# END Class GuiDocEditHeader

# =============================================================================================== #
#  The Embedded Document Footer
#  Only used by DocEditor, and is at a fixed position in the QTextEdit's viewport
# =============================================================================================== #

class GuiDocEditFooter(QWidget):

    def __init__(self, docEditor):
        QWidget.__init__(self, docEditor)

        logger.debug("Initialising GuiDocEditFooter ...")

        self.mainConf   = nw.CONFIG
        self.docEditor  = docEditor
        self.theParent  = docEditor.theParent
        self.theProject = docEditor.theProject
        self.theTheme   = docEditor.theTheme
        self.optState   = docEditor.theProject.optState
        self.theHandle  = None
        self.theItem    = None

        self.sPx = int(round(0.9*self.theTheme.baseIconSize))
        fPx = int(0.9*self.theTheme.fontPixelSize)
        bSp = self.mainConf.pxInt(4)
        hSp = self.mainConf.pxInt(6)

        lblFont = self.font()
        lblFont.setPointSizeF(0.9*self.theTheme.fontPointSize)

        # Main Widget Settings
        self.setContentsMargins(0, 0, 0, 0)
        self.setAutoFillBackground(True)

        # Status
        self.statusIcon = QLabel("")
        self.statusIcon.setContentsMargins(0, 0, 0, 0)
        self.statusIcon.setFixedHeight(self.sPx)
        self.statusIcon.setAlignment(Qt.AlignLeft | Qt.AlignTop)

        self.statusText = QLabel("Status")
        self.statusText.setIndent(0)
        self.statusText.setMargin(0)
        self.statusText.setContentsMargins(0, 0, 0, 0)
        self.statusText.setAutoFillBackground(True)
        self.statusText.setFixedHeight(fPx)
        self.statusText.setAlignment(Qt.AlignLeft | Qt.AlignTop)
        self.statusText.setFont(lblFont)

        # Lines
        self.linesIcon = QLabel("")
        self.linesIcon.setPixmap(self.theTheme.getPixmap("status_lines", (self.sPx, self.sPx)))
        self.linesIcon.setContentsMargins(0, 0, 0, 0)
        self.linesIcon.setFixedHeight(self.sPx)
        self.linesIcon.setAlignment(Qt.AlignLeft | Qt.AlignTop)

        self.linesText = QLabel("Line: 0")
        self.linesText.setIndent(0)
        self.linesText.setMargin(0)
        self.linesText.setContentsMargins(0, 0, 0, 0)
        self.linesText.setAutoFillBackground(True)
        self.linesText.setFixedHeight(fPx)
        self.linesText.setAlignment(Qt.AlignLeft | Qt.AlignTop)
        self.linesText.setFont(lblFont)

        # Words
        self.wordsIcon = QLabel("")
        self.wordsIcon.setPixmap(self.theTheme.getPixmap("status_stats", (self.sPx, self.sPx)))
        self.wordsIcon.setContentsMargins(0, 0, 0, 0)
        self.wordsIcon.setFixedHeight(self.sPx)
        self.wordsIcon.setAlignment(Qt.AlignLeft | Qt.AlignTop)

        self.wordsText = QLabel("Words: 0")
        self.wordsText.setIndent(0)
        self.wordsText.setMargin(0)
        self.wordsText.setContentsMargins(0, 0, 0, 0)
        self.wordsText.setAutoFillBackground(True)
        self.wordsText.setFixedHeight(fPx)
        self.wordsText.setAlignment(Qt.AlignLeft | Qt.AlignTop)
        self.wordsText.setFont(lblFont)

        # Assemble Layout
        self.outerBox = QHBoxLayout()
        self.outerBox.setSpacing(bSp)
        self.outerBox.addWidget(self.statusIcon)
        self.outerBox.addWidget(self.statusText)
        self.outerBox.addStretch(1)
        self.outerBox.addWidget(self.linesIcon)
        self.outerBox.addWidget(self.linesText)
        self.outerBox.addSpacing(hSp)
        self.outerBox.addWidget(self.wordsIcon)
        self.outerBox.addWidget(self.wordsText)
        self.setLayout(self.outerBox)

        # Fix the Colours
        self.matchColours()

        logger.debug("GuiDocEditFooter initialisation complete")

        return

    ##
    #  Methods
    ##

    def matchColours(self):
        """Update the colours of the widget to match those of the syntax
        theme rather than the main GUI.
        """
        thePalette = QPalette()
        thePalette.setColor(QPalette.Window, QColor(*self.theTheme.colBack))
        thePalette.setColor(QPalette.WindowText, QColor(*self.theTheme.colText))
        thePalette.setColor(QPalette.Text, QColor(*self.theTheme.colText))

        self.setPalette(thePalette)
        self.statusText.setPalette(thePalette)
        self.linesText.setPalette(thePalette)
        self.wordsText.setPalette(thePalette)

        return

    def setHandle(self, tHandle):
        """Set the handle that will populate the footer's data.
        """
        self.theHandle = tHandle
        if self.theHandle is None:
            logger.verbose("No handle set, so clearing the editor footer")
            self.theItem = None
        else:
            self.theItem = self.theProject.projTree[self.theHandle]

        self.updateInfo()
        self.updateCounts()

        return

    def updateInfo(self):
        """Update the content of text labels.
        """
        if self.theItem is None:
            sIcon = QPixmap()
            sText = ""
        else:
            iStatus = self.theItem.itemStatus
            if self.theItem.itemClass == nwItemClass.NOVEL:
                iStatus = self.theProject.statusItems.checkEntry(iStatus)
                theIcon = self.theParent.statusIcons[iStatus]
            else:
                iStatus = self.theProject.importItems.checkEntry(iStatus)
                theIcon = self.theParent.importIcons[iStatus]
            sIcon = theIcon.pixmap(self.sPx, self.sPx)
            sText = self.theItem.itemStatus

        self.statusIcon.setPixmap(sIcon)
        self.statusText.setText(sText)

        return

    def updateLineCount(self):
        """Update the word count.
        """
        if self.theItem is None:
            iLine = 0
            iDist = 0
        else:
            theCursor = self.docEditor.textCursor()
            iLine = theCursor.blockNumber() + 1
            iDist = 100*iLine/self.docEditor.qDocument.blockCount()

        self.linesText.setText(f"Line: {iLine:n} ({iDist:.0f}\u202f%)")

        return

    def updateCounts(self):
        """Update the word count.
        """
        if self.theItem is None:
            wCount = 0
            wDiff  = 0
        else:
            wCount = self.theItem.wordCount
            wDiff  = wCount - self.theItem.initCount

        self.wordsText.setText(f"Words: {wCount:n} ({wDiff:+n})")

        byteSize = self.docEditor.qDocument.characterCount()
        self.wordsText.setToolTip(f"Document size is {byteSize:n} bytes")

        return

# END Class GuiDocEditFooter
